//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_Action
// Author      : BehemothProgrammer
// Description : Executes a script action when triggered
// OnTrigger   : Runs the script
//
// Parameters  : (7) Params 1 -128..127					: Action Index
//               (3) Max Regenerations -32768..32767	: Flags
//               Radius									: Delay before executing script action
//               WallRadius								: Arg 0
//               Height									: Arg 1
//               StepHeight/DeadHeight					: Arg 2
//               (2) Health -32768..32767				: Arg 3
//               (0) Target TID -32768..32767 			: Arg 4 (Common: Actor TID)
//               (6) Params 2 -128..127					: Arg 5
//               (4) Attach Chance 0..255				: Arg 6
//               (5) Texture -128..127					: Arg 7
//               (1) Trigger Anim 0..255				: Arg 8
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum ActionFlags
	{
		ACTFLAG_ONBEGINLEVEL = (1 << 0), //if is marked persistent OnRestore then will execute script action OnPostBeginLevel
		ACTFLAG_REPEATABLE = (1 << 1),
		ACTFLAG_NOPERSIST = (1 << 2) //Will never mark as persistent
	}

	enum ActionList
	{
		ACTION_ACTOR_REMOVE = 0, //Removes actor, Arg 4 = Actor TID
		ACTION_ACTOR_AISHOW = 1, //Show AI actor, Arg 4 = Actor TID. Use spawn flag1 "Respawn Animation" to show a spawn effect when the ai is shown.
		ACTION_ACTOR_AIHIDE = 2, //Hide AI actor, Arg 4 = Actor TID
		ACTION_ACTOR_SOLID_NONACTIVATED = 3, //Turns On actors Solid flag if actor is not activated
		ACTION_ACTOR_TAKEINVENTORY = 4, //Take x amount of inventory. Arg 5 = Inventory Index 0..127, Arg 6 = Amount to take
		ACTION_ACTOR_GIVEINVENTORY = 5, //Give x amount of inventory. Arg 5 = Inventory Index 0..127, Arg 6 = Amount to take
		ACTION_ACTOR_HURTPLAYER = 6, //Inflict damage on the player. Arg 3 = Damage
		ACTION_ACTOR_AITRIGGER = 7, //Triggers AI for use with their Trigger Anim. Arg 4 = Actor TID.
	}

	enum ActionStates
	{
		ACTSTATE_IDLE = 0,
		ACTSTATE_DELAYING
	}

	class CL_BP_Action : CL_BP_ScriptObject
	{
		int flags;
		array<int> args(9);
		float delay;
		int actionIndex;
		
		bool restoreCalled;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_Action(kActor @actor)
		{
			super(actor);
			state = ACTSTATE_IDLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			flags = self.SpawnParams(3);
			delay = self.Radius();
			actionIndex = self.SpawnParams(7);
			args[0] = int(self.WallRadius());
			args[1] = int(self.Height());
			args[2] = int(self.StepHeight());
			args[3] = self.SpawnParams(2);
			args[4] = self.SpawnParams(0);
			args[5] = self.SpawnParams(6);
			args[6] = self.SpawnParams(4);
			args[7] = self.SpawnParams(5);
			args[8] = self.SpawnParams(1);
			
			SetState(ACTSTATE_IDLE);
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			restoreCalled = true;
			self.Flags() |= AF_ACTIVATED;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnPostBeginLevel()
		{
			if (restoreCalled && flags & ACTFLAG_ONBEGINLEVEL != 0)
			{
				DoAction(true);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			switch (state)
			{
				case ACTSTATE_IDLE:
				{
					break;
				}
				case ACTSTATE_DELAYING:
				{
					if (BP::Time::Reached(stateEndTime))
					{
						DoAction();
						SetState(ACTSTATE_IDLE);
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			if (flags & ACTFLAG_REPEATABLE != 0)
			{
				self.Flags() &= ~AF_ACTIVATED;
				if (state != ACTSTATE_IDLE)
					return;
			}
			else
			{
				if (flags & ACTFLAG_NOPERSIST == 0)
					self.MarkPersistentBit(false);
			}
			
			if (delay > 0.0f)
				SetState(ACTSTATE_DELAYING);
			else
				DoAction();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			switch (state)
			{
				case ACTSTATE_IDLE:
				{
					self.Flags() &= ~AF_ALWAYSACTIVE;
					break;
				}
				case ACTSTATE_DELAYING:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					stateEndTime = BP::Time::FromNow(delay);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoAction(const bool onLevelStart = false)
		{
			switch (actionIndex)
			{
				case ACTION_ACTOR_REMOVE:
				{
					int actorTID = args[4];
					kActor@ actor = World.GetActorByTID(actorTID);
					if (BP::Actor::Exists(actor))
						actor.Remove();
					break;
				}
				case ACTION_ACTOR_AISHOW:
				{
					int actorTID = args[4];
					kActor@ actor = World.GetActorByTID(actorTID);
					if (BP::Actor::Exists(actor))
					{
						BP::Actor::ShowAI(actor);
						if (!onLevelStart && self.SpawnFlags1() & BP::Math::UBIT31 != 0) //respawn animation
						{
							actor.PlaySound("sounds/shaders/high_priest_teleport.ksnd");
							actor.SpawnFx("fx/regeneration_appearance.kfx", Math::vecZero);
							actor.SpawnFx("fx/regeneration_start.kfx", Math::vecZero);
						}
					}
					break;
				}
				case ACTION_ACTOR_AIHIDE:
				{
					int actorTID = args[4];
					kActor@ actor = World.GetActorByTID(actorTID);
					if (BP::Actor::Exists(actor))
						BP::Actor::HideAI(actor);
					break;
				}
				case ACTION_ACTOR_SOLID_NONACTIVATED:
				{
					int actorTID = args[4];
					kActor@ actor = World.GetActorByTID(actorTID);
					if (BP::Actor::Exists(actor) && !BP::Actor::IsActivated(actor))
						actor.Flags() |= AF_SOLID;
					break;
				}
				case ACTION_ACTOR_TAKEINVENTORY:
				{
					int inventoryIndex = args[5];
					int amount = args[6];
					BP::Inventory::Take(inventoryIndex, amount);
					break;
				}
				case ACTION_ACTOR_GIVEINVENTORY:
				{
					int inventoryIndex = args[5];
					int amount = args[6];
					BP::Inventory::Give(inventoryIndex, amount);
					break;
				}
				case ACTION_ACTOR_HURTPLAYER:
				{
					int damage = args[3];
					playerActor.InflictGenericDamage(self, damage);
					break;
				}
				case ACTION_ACTOR_AITRIGGER:
				{
					int actorTID = args[4];
					kActor@ actor = World.GetActorByTID(actorTID);
					if (BP::Actor::Exists(actor))
					{
						actor.Flags() &= ~(AF_DISABLED | AF_INVINCIBLE);
						World.TriggerActorsByTID(self, actorTID);
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
