//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_ActorSpawner
// Author      : BehemothProgrammer
// Description : Spawns an actor in a radius or in a polygon of points. With min/max respawn times and other flag options.
// OnTrigger   : Toggles Enable/Disable the spawner. Removes all existing spawned objects when disabled. Is persistent.
//
// Parameters  : (7) Params 1 -128..127					: Spawn Type
//               (6) Params 2 -128..127					: Max number of spawns. Infinite amount if less than 1.
//               (3) Max Regenerations -32768..32767	: Flags
//               (2) Health -32768..32767				: ActorID to spawn
//               (0) Target TID -32768..32767 			: Amount of actors to create per spawn (always at least 1)
//               Radius									: Circle Spawn Radius / Polygon next actorTID to use as point (uses that actors Target TID for next TID until is less than 1)
//               WallRadius								: Currently not used.
//               Height 								: Respawn Min Time
//               StepHeight/DeadHeight					: Respawn Max Time
//               Scale									: Scale for each spawned actor
//               (1) Trigger Anim 0..255				: Min scale factor if has ACTORSPAWNER_FLAG_RANDSCALE flag (value is divided 10)
//               (4) Attach Chance 0..255				: Max scale factor if has ACTORSPAWNER_FLAG_RANDSCALE flag (value is divided 10)
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum ActorSpawnerTypes
	{
		ACTORSPAWNER_TYPE_RADIUS = 0,
		ACTORSPAWNER_TYPE_POLYGON = 1
	}

	enum ActorSpawnerFlags
	{
		ACTORSPAWNER_FLAG_SPAWNONACTIVATE = (1 << 0),
		ACTORSPAWNER_FLAG_STARTDISABLED = (1 << 1),
		ACTORSPAWNER_FLAG_SPARKLES = (1 << 2),
		ACTORSPAWNER_FLAG_SPAWNFX = (1 << 3),
		ACTORSPAWNER_FLAG_RANDSCALE = (1 << 4),
		ACTORSPAWNER_FLAG_RANDYAW = (1 << 5),
		ACTORSPAWNER_FLAG_FLOORHEIGHT = (1 << 6) //use floorheight for z position of spawned actors
	}

	class CL_BP_ActorSpawner : CL_BP_ScriptObject
	{
		int spawnType;
		int flags;
		int spawnActorID;
		float spawnMinTime;
		float spawnMaxTime;
		int actorsPerSpawn;
		int maxSpawnCount;
		float scaleFactorMin;
		float scaleFactorMax;
		float spawnRadius;
		
		bool isActive;
		int spawnCount;
		float spawnTime;
		CL_BP_ActorIterator actorsIter;
		array<kVec3> polyPoints;
		array<int> polyIndices;
		kSelectionListInt polyTriList;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_ActorSpawner(kActor @actor)
		{
			super(actor);
			isActive = false;
			spawnTime = 0.0f;
			spawnCount = 0;
			actorsIter.Setup(10, BP_FUNCDEF_ACTORITERTICK(OnActorTick), BP_FUNCDEF_ACTORITEREMPTY(OnActorsEmpty));
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			spawnType = self.SpawnParams(7);
			maxSpawnCount = self.SpawnParams(6);
			flags = self.SpawnParams(3);
			spawnActorID = self.SpawnParams(2);
			actorsPerSpawn = BP::Math::Max(self.SpawnParams(0), 1);
			spawnMinTime = self.Height();
			spawnMaxTime = self.StepHeight();
			scaleFactorMin = (self.SpawnParams(1) <= 0) ? 1.0f : float(self.SpawnParams(1)) / 10.0f;
			scaleFactorMax = (self.SpawnParams(4) <= 0) ? 1.0f : float(self.SpawnParams(4)) / 10.0f;
			spawnRadius = self.Radius();
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnPostBeginLevel()
		{
			//For polygon spawner get the actors as points on the map
			if (spawnType == ACTORSPAWNER_TYPE_POLYGON)
			{
				polyPoints.insertLast(self.Origin());
				int nextTID = int(spawnRadius);
				while (nextTID > 0)
				{
					kActor@ pointActor = World.GetActorByTID(nextTID);
					if (BP::Actor::Exists(pointActor))
					{
						polyPoints.insertLast(pointActor.Origin());
						nextTID = pointActor.SpawnParams(0);
					}
					else
					{
						break;
					}
				}
				polyPoints.reverse();
				
				polyIndices = BP::Math::Triangulate(polyPoints);
				for (uint i = 0; i < polyIndices.length(); i+=3)
					polyTriList.AddItem(i, int(BP::Math::AreaOfTriangle(polyPoints[polyIndices[i]], polyPoints[polyIndices[i+1]], polyPoints[polyIndices[i+2]])));
			}

			//Set initial active
			bool startActive = !self.IsPersistentMarked();
			if (flags & ACTORSPAWNER_FLAG_STARTDISABLED != 0)
				startActive = !startActive;
			SetActive(startActive);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			actorsIter.Tick();
			TrySpawn();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TrySpawn()
		{
			// if (self.GameTicks() % 40 == 0)
				// Log("Active: " + isActive + " iterEmpty: " + actorsIter.IsEmpty() + " ReachedSpawnLimit: " + ReachedSpawnLimit() + " TimeReached: " + BP::Time::Reached(spawnTime) + " GameTime: " + BP::Game::Time() + " spawnTime: " + spawnTime);
			if (!isActive || !actorsIter.IsEmpty() || ReachedSpawnLimit() || !BP::Time::Reached(spawnTime))
				return;
			
			SpawnActors();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			self.Flags() &= ~AF_ACTIVATED;
			spawnCount = 0;
			BP::Actor::TogglePersistentBit(self);
			bool setActive = !self.IsPersistentMarked();
			if (flags & ACTORSPAWNER_FLAG_STARTDISABLED != 0)
				setActive = !setActive;
			SetActive(setActive);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActorsEmpty()
		{
			SetNextSpawnTime();
		}		
		//------------------------------------------------------------------------------------------------------------------------
		bool OnActorTick(kActor@ actor)
		{
			if (flags & ACTORSPAWNER_FLAG_SPARKLES != 0 && actor.GameTicks() % 2 == 0)
				actor.SpawnFx("fx/generic_230.kfx", kVec3(0, 0, GAME_SCALE));
			
			return !BP::Actor::Exists(actor);
		}		
		//------------------------------------------------------------------------------------------------------------------------
		void SpawnActors()
		{
			SetNextSpawnTime();
			kActor@ actor = null;
			float yaw = self.Yaw();
			spawnCount++;
			
			for (int i = 0; i < actorsPerSpawn; i++)
			{
				//Yaw
				if (flags & ACTORSPAWNER_FLAG_RANDYAW != 0)
					yaw = Math::RandRange(-Math::pi, Math::pi);
				
				//Position
				kVec3 spawnPos;
				if (spawnType == ACTORSPAWNER_TYPE_RADIUS)
				{
					spawnPos = BP::Math::RandPointInCircle(self.Origin(), spawnRadius);
				}
				else if (spawnType == ACTORSPAWNER_TYPE_POLYGON)
				{
					int triIndice = polyTriList.Select();
					spawnPos = BP::Math::RandPointInTriangle(polyPoints[polyIndices[triIndice]], polyPoints[polyIndices[triIndice+1]], polyPoints[polyIndices[triIndice+2]]);
				}
				
				int spawnSector = self.GetSectorIndexAtLocation(spawnPos);
				@actor = BP::Actor::Spawn(spawnActorID, spawnPos, yaw, spawnSector);
				if (BP::Actor::Exists(actor))
				{
					actorsIter.AddActor(actor);
					if (flags & ACTORSPAWNER_FLAG_FLOORHEIGHT != 0)
						actor.Origin().z = actor.FloorHeight();
					actor.Scale() = self.Scale();
					if (flags & ACTORSPAWNER_FLAG_RANDSCALE != 0)
					{
						float scaleFactor = Math::RandRange(scaleFactorMin, scaleFactorMax);
						actor.Scale() *= scaleFactor;
						actor.Radius() *= scaleFactor;
						actor.Height() *= scaleFactor;
					}
					
					//don't do spawn effects/sound if spawned at start
					if (flags & ACTORSPAWNER_FLAG_SPAWNFX != 0 && self.GameTicks() != 0)
					{
						kVec3 org = actor.Origin();
						org.z += (3 * GAME_SCALE);
						actor.PlaySound("sounds/shaders/generic_3_energy_pickup.ksnd");
						Game.SpawnFx("fx/generic_260.kfx", org, actor.SectorIndex());
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetActive(const bool active)
		{
			isActive = active;
			if (active)
			{
				if (flags & ACTORSPAWNER_FLAG_SPAWNONACTIVATE != 0)
					spawnTime = BP::Time::FromNow(0.0f);
				else
					SetNextSpawnTime();
				
				TrySpawn();
			}
			else
			{
				actorsIter.Clear();
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetNextSpawnTime()
		{
			spawnTime = BP::Time::FromNow(Math::RandRange(spawnMinTime, spawnMaxTime));
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool ReachedSpawnLimit()
		{
			return maxSpawnCount > 0 && spawnCount >= maxSpawnCount;
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
