//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_Counter
// Author      : BehemothProgrammer
// Description : Increases an integer counter each time it's triggered. Once it's limit is reached it can trigger actors and scripts.
//               Counter values are persistent by use of GameVariables.
//				 if marked persistent then will trigger actors/scripts in OnPostBeginLevel if CF_NOPERSISTCALL is not on.
//
// OnTrigger   : Increases counter value by 1. When the limit is reached the persistent bit is set and subsequent triggers do nothing.
//
// Parameters  : (7) Params 1 -128..127 				: Counter ID 0..127
//               Radius									: Counter Limit
//               WallRadius								: Amount to increase counter when triggered.
//														  If CF_BITMASKON/OFF then is the bit to turn on/off (0..31) negative values turn it off.
//               (3) Max Regenerations -32768..32767	: flags
//               (0) Target TID -32768..32767 			: Actors TID to trigger when counter has reached it's limit. If is less than 1 this will not be used.
//               (2) Health -32768..32767				: Script to trigger when counter has reached it's limit. If is less than 1 this will not be used.
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum CounterFlags
	{
		CF_NOPERSISTCALL = (1 << 0), //Will not trigger on map start when marked persistent
		CF_TRIGGEREVERYTIME = (1 << 1), //Will trigger actors/scripts every time this counter is triggered instead of only when the limit is reached.
		CF_BITMASKON = (1 << 2), //uses the amount to increase as the bit to turn on
		CF_BITMASKOFF = (1 << 3), //uses the amount to increase as the bit to turn off
		CF_LIMITEQUAL = (1 << 4), //value must be equal to the limit to trigger instead of >=
		CF_LIMITLESSEQUAL = (1 << 5) //value must be Less than or equal to the limit to trigger instead of >=
	}

	class CL_BP_Counter : CL_BP_ScriptObject
	{
		int id;
		int limit;
		int flags;
		int incAmount;
		int triggerTID;
		int triggerScriptID;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_Counter(kActor @actor)
		{
			super(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			id = self.SpawnParams(7);
			limit = int(self.Radius());
			incAmount = int(self.WallRadius());
			flags = self.SpawnParams(3);
			triggerTID = self.SpawnParams(0);
			triggerScriptID = self.SpawnParams(2);
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			self.Flags() |= AF_ACTIVATED;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnPostBeginLevel()
		{
			if (self.IsPersistentMarked() && flags & ELF_NOPERSISTCALL == 0)
				TriggerActors();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			if (id < 0 || id >= MAX_COUNTERS)
				return;
			
			int value = 0;
			if (flags & CF_BITMASKON != 0)
				value = BP::Game::SetCounterBitMask(id, (1 << incAmount), true);
			else if (flags & CF_BITMASKOFF != 0)
				value = BP::Game::SetCounterBitMask(id, (1 << incAmount), false);
			else
				value = BP::Game::IncreaseCounter(id, incAmount);
			
			bool hitLimit = false;
			if (flags & CF_LIMITEQUAL != 0)
				hitLimit = value == limit;
			else if (flags & CF_LIMITLESSEQUAL != 0)
				hitLimit = value <= limit;
			else
				hitLimit = value >= limit;
			
			if (hitLimit)
			{
				self.MarkPersistentBit(false);
				TriggerActors();
			}
			else
			{
				if (flags & CF_TRIGGEREVERYTIME != 0)
					TriggerActors();
				self.Flags() &= ~AF_ACTIVATED;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TriggerActors()
		{
			if (triggerTID > 0)
				World.TriggerActorsByTID(self, triggerTID);
			if (triggerScriptID > 0)
				Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
