//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_EventLink
// Author      : BehemothProgrammer
// Description : Triggers actors/script when condition is met also when restored unless flagged not to.
//               Chain multiple EventLinks together for more conditions.
//               One main use for EventLinks is to keep enemies that were activated and alive on the map still active next time the level is loaded.
// OnTrigger   : If condition is met then it will trigger target actors and/or call scriptID.
//
// Parameters  : (7) Params 1 -128..127					: Condition
//               (6) Params 2 -128..127					: Condition Arg 1
//               (4) Attach Chance 0..255				: Condition Arg 2
//               Height									: Condition Arg 3
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: Actors TID to trigger. If is less than 1 this will not be used.
//               (2) Health -32768..32767				: Script to trigger. If is less than 1 this will not be used.
//               Radius									: Delay before triggering actors/script
//               WallRadius								: Amount of time to wait until can be triggered again (If is repeatable)
//               StepHeight/DeadHeight					: 
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum EventLinkFlags
	{
		ELF_REPEATABLE = (1 << 0), //Triggers every OnActivate when condition is met instead of once. Is never marked persistent.
		ELF_INVERSECONDITION = (1 << 1),
		ELF_NOPERSISTCALL = (1 << 2), //Will not trigger on map start when marked persistent
		ELF_NOPERSIST = (1 << 3) //Will never mark as persistent
	}

	enum EventLinkConditions
	{
		ELC_NONE = 0,
		ELC_DIFFICULTY = 1, //Arg1 is the difficulty flags that must be on (BPEventLinkConditionDifficultyFlags)
		ELC_PLAYER_GROUNDED = 2, //Player must be grounded
		ELC_INVENTORY = 3, //Has Inventory Item. Arg1 is the item index. Arg2 is the amount required to have.
		ELC_KEY = 4, //Has collected key. Arg1 is the hubID (normally the first keys are for hub 2). Arg2 is the keybit.
		ELC_ACTOR_ACTIVATED = 5, //Actor is activated. Arg3 is the TID of the actor.
		ELC_ACTOR_DEAD = 6, //Actor is Dead or doesn't exist. Arg3 is the TID of the actor.
		ELC_COUNTER = 7, //Counter is a certain value. arg1 is CounterID. arg2 is operator. Arg3 is counter value.
		ELC_PLAYER_GROUNDEDWATER = 8, //Player must be grounded or in water
		ELC_PLAYER_INACTORAREA = 9, //Player must be inside the actors collision cylinder. Arg3 is the TID of the actor.
	}

	enum EventLinkCounterOperators
	{
		ELC_OP_EQUALS = 0,
		ELC_OP_GREATEREQUALS = 1,
		ELC_OP_LESSEQUALS = 2,
		ELC_OP_GREATER = 3,
		ELC_OP_LESS = 4,
	}

	enum EventLinkConditionDifficultyFlags
	{
		ELC_DF_EASY = (1 << 0),
		ELC_DF_NORMAL = (1 << 1),
		ELC_DF_HARD = (1 << 2),
		ELC_DF_HARDCORE = (1 << 3)
	}

	enum EventLinkStates
	{
		ELS_IDLE = 0,
		ELS_DELAYING,
		ELS_RETRIGGERDELAY,
	}

	class CL_BP_EventLink : CL_BP_ScriptObject
	{
		int condition;
		int conditionArg1;
		int conditionArg2;
		int conditionArg3;
		int flags;
		int triggerTID;
		int triggerScriptID;
		float delay;
		float retriggerDelay;
		
		kActor@ conditionActor;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_EventLink(kActor @actor)
		{
			super(actor);
			state = ELS_IDLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			condition = self.SpawnParams(7);
			conditionArg1 = self.SpawnParams(6);
			conditionArg2 = self.SpawnParams(4);
			conditionArg3 = int(self.Height());
			flags = self.SpawnParams(3);
			triggerTID = self.SpawnParams(0);
			triggerScriptID = self.SpawnParams(2);
			delay = self.Radius();
			retriggerDelay = self.WallRadius();
			
			switch (condition)
			{
				case ELC_ACTOR_ACTIVATED:
				case ELC_ACTOR_DEAD:
				case ELC_PLAYER_INACTORAREA:
					@conditionActor = World.GetActorByTID(conditionArg3);
					break;
			}
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			self.Flags() |= AF_ACTIVATED;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnPostBeginLevel()
		{
			if (self.IsPersistentMarked() && flags & ELF_NOPERSISTCALL == 0)
				TriggerLink();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			switch (state)
			{
				case ELS_IDLE:
				{
					break;
				}
				case ELS_DELAYING:
				{
					if (BP::Time::Reached(stateEndTime))
						TriggerLink();
					break;
				}
				case ELS_RETRIGGERDELAY:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(ELS_IDLE);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			if (!IsConditionMet())
			{
				self.Flags() &= ~AF_ACTIVATED;
				return;
			}
			
			ScheduleTriggerLink();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			switch (state)
			{
				case ELS_IDLE:
				{
					self.Flags() &= ~AF_ALWAYSACTIVE;
					if (flags & ELF_REPEATABLE != 0)
						self.Flags() &= ~AF_ACTIVATED;
					break;
				}
				case ELS_DELAYING:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					stateEndTime = BP::Time::FromNow(delay);
					break;
				}
				case ELS_RETRIGGERDELAY:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					stateEndTime = BP::Time::FromNow(retriggerDelay);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ScheduleTriggerLink()
		{
			if (flags & ELF_REPEATABLE == 0 && flags & ELF_NOPERSIST == 0)
				self.MarkPersistentBit(false);
			
			if (delay <= 0.0f)
			{
				TriggerLink();
				return;
			}
			
			SetState(ELS_DELAYING);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TriggerLink()
		{
			// Log("EventLink " + self.TID() + " Triggered TID " + triggerTID + " and Script " + triggerScriptID);
			
			if (retriggerDelay <= 0.0f || flags & ELF_REPEATABLE == 0)
				SetState(ELS_IDLE);
			else
				SetState(ELS_RETRIGGERDELAY);
			
			if (flags & ELF_REPEATABLE == 0 && flags & ELF_NOPERSIST == 0)
				self.MarkPersistentBit(false);
			
			if (triggerTID > 0 && triggerTID != self.TID()) //uses actor TID and not triggering self again
				World.TriggerActorsByTID(self, triggerTID);
			if (triggerScriptID > 0)
				Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool IsConditionMet()
		{
			bool result = true;
			switch (condition)
			{
				case ELC_DIFFICULTY:
				{
					int difficulty = Game.GetDifficulty();
					result = ((difficulty == DIFFICULTY_EASY && conditionArg1 & ELC_DF_EASY != 0) ||
							(difficulty == DIFFICULTY_NORMAL && conditionArg1 & ELC_DF_NORMAL != 0) ||
							(difficulty == DIFFICULTY_HARD && conditionArg1 & ELC_DF_HARD != 0) ||
							(difficulty == DIFFICULTY_HARDCORE && conditionArg1 & ELC_DF_HARDCORE != 0));
					break;
				}
				case ELC_PLAYER_GROUNDED:
				{
					result = Player.Actor().OnGround();
					break;
				}
				case ELC_INVENTORY:
				{
					result = BP::Inventory::Count(conditionArg1) >= conditionArg2;
					break;
				}
				case ELC_KEY:
				{
					result = BP::Inventory::HasKey(conditionArg1, conditionArg2);
					break;
				}
				case ELC_ACTOR_ACTIVATED:
				{
					result = BP::Actor::Exists(conditionActor) && BP::Actor::IsActivated(conditionActor);
					break;
				}
				case ELC_ACTOR_DEAD:
				{
					result = !BP::Actor::Exists(conditionActor) || BP::Actor::IsDead(conditionActor);
					break;
				}
				case ELC_COUNTER:
				{
					int counterID = conditionArg1;
					int op = conditionArg2;
					int targetValue = conditionArg3;
					int counterValue = BP::Game::GetCounter(counterID);
					if (op == ELC_OP_EQUALS)
						result = counterValue == targetValue;
					else if (op == ELC_OP_GREATEREQUALS)
						result = counterValue >= targetValue;
					else if (op == ELC_OP_LESSEQUALS)
						result = counterValue <= targetValue;
					else if (op == ELC_OP_GREATER)
						result = counterValue > targetValue;
					else if (op == ELC_OP_LESS)
						result = counterValue < targetValue;
					break;
				}
				case ELC_PLAYER_GROUNDEDWATER:
				{
					int waterLevel = playerActor.GetWaterLevel();
					result = (Player.Actor().OnGround() || waterLevel == WLT_BETWEEN || waterLevel == WLT_UNDER);
					break;
				}
				case ELC_PLAYER_INACTORAREA:
				{
					result = BP::Actor::Exists(conditionActor) && BP::Actor::InPoint(conditionActor, playerActor.Origin());
					break;
				}
			}
			
			return (flags & ELF_INVERSECONDITION != 0) ? !result : result;
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
