//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_FxSpawner
// Author      : BehemothProgrammer
// Description : Spawns an Fx with a delay and other flag options.
// OnTrigger   : Spawns the fx after delay. If OneShot flag is on then will not spawn again and is persistent.
//
// Parameters  : (3) Max Regenerations -32768..32767	: Flags
//               (2) Health -32768..32767				: Fx to spawn (ie. 12 = explosion)
//               (0) Target TID -32768..32767 			: Sound to play when fx spawns. If is less than 1 this will not be used. (ie. 66 = explosion_2)
//               Radius									: Delay before spawning fx
//               Wall Radius							: Delay until stop spawning fx (if <= 0 then only spawns once)
//               Height									: Delay until can be triggered again after stopped spawning
//               StepHeight/DeadHeight					: Time between each fx spawn while in spawning state
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum FxSpawnerFlags
	{
		FXSPAWNER_FLAG_CUSTOM = (1 << 0), //use custom fx from BP_GameDefCustom actor def
		FXSPAWNER_FLAG_ONESHOT = (1 << 1), //plays only once and sets persistent bit
		FXSPAWNER_FLAG_SPAWNONSTART = (1 << 2), //spawns automatically on start
		FXSPAWNER_FLAG_ALWAYSTRIGGER = (1 << 3), //when entering idle state will start spawning fx (going forever)
		FXSPAWNER_FLAG_CUSTOMSOUND = (1 << 4) //use custom sound when fx spawns from BP_GameDefCustom actor def
	}

	enum FxSpawnerStates
	{
		FXSS_IDLE = 0,
		FXSS_DELAY2SPAWNING,
		FXSS_SPAWNING,
		FXSS_DELAY2IDLE
	}

	class CL_BP_FxSpawner : CL_BP_ScriptObject
	{
		int flags;
		int fxID;
		int soundID;
		float startDelay;
		float offDelay;
		float retriggerDelay;
		float spawnTimeDelay;
		
		float spawnTime;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_FxSpawner(kActor @actor)
		{
			super(actor);
			state = FXSS_IDLE;
			spawnTime = 0.0f;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			flags = self.SpawnParams(3);
			fxID = self.SpawnParams(2);
			soundID = self.SpawnParams(0);
			startDelay = BP::Math::Max(self.Radius(), 0.0f);
			offDelay = BP::Math::Max(self.WallRadius(), 0.0f);
			retriggerDelay = BP::Math::Max(self.Height(), 0.0f);
			spawnTimeDelay = BP::Math::Max(self.StepHeight(), 0.0f);
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
			
			SetState(FXSS_IDLE);
			
			if (self.IsPersistentMarked())
			{
				self.Flags() |= AF_ACTIVATED;
			}
			else
			{
				if (flags & FXSPAWNER_FLAG_SPAWNONSTART != 0)
					StartSpawnFx();
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			switch (state)
			{
				case FXSS_IDLE:
				{
					break;
				}
				case FXSS_DELAY2SPAWNING:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(FXSS_SPAWNING);
					break;
				}
				case FXSS_SPAWNING:
				{
					if (BP::Time::Reached(spawnTime))
					{
						spawnTime = BP::Time::FromNow(spawnTimeDelay);
						if (soundID > 0)
						{
							if (flags & FXSPAWNER_FLAG_CUSTOMSOUND != 0)
								self.PlaySound(BP::Game::GetCustomSound(soundID));
							else
								self.PlaySoundWithLookupID(soundID);
						}
						if (BP::Game::HasFx(fxID))
						{
							if (flags & FXSPAWNER_FLAG_CUSTOM != 0)
								self.SpawnFx(BP::Game::GetCustomFx(fxID), Math::vecZero, self.Rotation());
							else
								self.SpawnFx(BP::Game::GetFx(fxID), Math::vecZero, self.Rotation());
						}
					}
					
					if (BP::Time::Reached(stateEndTime))
					{
						if (retriggerDelay <= 0.0f)
							SetState(FXSS_IDLE);
						else
							SetState(FXSS_DELAY2IDLE);
					}
					break;
				}
				case FXSS_DELAY2IDLE:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(FXSS_IDLE);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			self.Flags() &= ~AF_ACTIVATED;
			StartSpawnFx();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			switch (state)
			{
				case FXSS_IDLE:
				{
					self.StopSound();
					self.Flags() &= ~AF_ALWAYSACTIVE;
					
					if (flags & FXSPAWNER_FLAG_ALWAYSTRIGGER != 0)
					{
						if (startDelay <= 0.0f)
							SetState(FXSS_SPAWNING);
						else
							SetState(FXSS_DELAY2SPAWNING);
					}
					break;
				}
				case FXSS_DELAY2SPAWNING:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					stateEndTime = BP::Time::FromNow(startDelay);
					break;
				}
				case FXSS_SPAWNING:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					spawnTime = 0.0f;
					stateEndTime = BP::Time::FromNow(offDelay);
					break;
				}
				case FXSS_DELAY2IDLE:
				{
					self.Flags() |= AF_ALWAYSACTIVE;
					self.Flags() &= ~AF_DORMANT;
					stateEndTime = BP::Time::FromNow(retriggerDelay);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void StartSpawnFx()
		{
			if (state != FXSS_IDLE)
				return;
			
			if (flags & FXSPAWNER_FLAG_ONESHOT != 0)
			{
				self.Flags() |= AF_ACTIVATED;
				self.MarkPersistentBit(false);
			}
			
			if (startDelay <= 0.0f)
				SetState(FXSS_SPAWNING);
			else
				SetState(FXSS_DELAY2SPAWNING);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
