//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_HUDMessages
// Author      : BehemothProgrammer
// Description : Shows messages when triggered
// OnTrigger   : Shows the message. If OneShot flag is on then will not show again and is persistent.
//
// Parameters  : (7) Params 1 -128..127 				: Message Type
//               (6) Params 2 -128..127 				: bitMask for which lines use NORMAL game localization text.
//														  DEFAULT bit (0) is Custom text from BP_GameDefCustom actor def is used by default.
//														  If HUDMSGTYPE_SHOWKEYSLEFT flag is on then is the hubID
//               (3) Max Regenerations -32768..32767	: Flags
//               (5) Texture -128..127                  : How long in frames the text will be visisble (1 = 60 frames) (default of 0 = 120 frames)
//														  If HUDMSG_FLAG_HELP flag is on then 60 frames is automatically added for the time the message takes to fade in.
//               Radius									: Text Line 1 (if < 0 then blank line)
//               WallRadius								: Text Line 2
//               Height									: Text Line 3
//               StepHeight/DeadHeight					: Text Line 4 
//               (2) Health -32768..32767				: Text Line 5 (soundID if flags HUDMSG_FLAG_SOUND or HUDMSG_FLAG_CUSTOMSOUND is on)
//               (0) Target TID -32768..32767 			: Text Line 6
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum HUDMessageTypes
	{
		HUDMSGTYPE_NORMAL = 0, //Shows each line value as Localized Text
		HUDMSGTYPE_SHOWKEYSLEFT = 1, //Shows the number of keys remaining for the hubID
		HUDMSGTYPE_PREPEND1 = 2, //prepends line 6 value to line 1
		HUDMSGTYPE_PREPEND2 = 3 //prepends line 6 counter index value to line 1
	}

	enum HUDMessageFlags
	{
		HUDMSG_FLAG_HELP = (1 << 0), //prints in a help box instead
		HUDMSG_FLAG_ONESHOT = (1 << 1), //plays only once and sets persistent bit
		HUDMSG_FLAG_SOUND = (1 << 2), //plays a sound when message is displayed. Text Line 5 is soundID.
		HUDMSG_FLAG_CUSTOMSOUND = (1 << 3), //plays a custom sound when message is displayed. Text Line 5 is soundID.
		HUDMSG_FLAG_ALIGN_CENTER = (1 << 4), //Aligns text for non help messages to the center instead of the top
	}

	class CL_BP_HUDMessages : CL_BP_ScriptObject
	{
		int msgType;
		int flags;
		int arg;
		int textVisibleTime;
		array<int> TextIDs(6);
		int hubID;
		int customTextMask; //bits that are on say that line should use the normal game localization and when off use custom text.
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_HUDMessages(kActor @actor)
		{
			super(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			msgType = int(self.SpawnParams(7));
			hubID = BP::Math::Max(int(self.SpawnParams(6)), 0);
			customTextMask = int(self.SpawnParams(6));
			flags = self.SpawnParams(3);
			textVisibleTime = self.SpawnParams(5);
			TextIDs[0] = int(self.Radius());
			TextIDs[1] = int(self.WallRadius());
			TextIDs[2] = int(self.Height());
			TextIDs[3] = int(self.StepHeight());
			TextIDs[4] = self.SpawnParams(2);
			TextIDs[5] = self.SpawnParams(0);
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			self.Flags() |= AF_ACTIVATED;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			if (flags & HUDMSG_FLAG_ONESHOT != 0)
			{
				self.MarkPersistentBit(false);
			}
			else
			{
				self.Flags() &= ~AF_ACTIVATED;
			}
			
			ShowMessages();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ShowMessages()
		{
			int language = BP::Game::GetCurrentLanguage();
			array<kStr> lines;
			switch (msgType)
			{
				case HUDMSGTYPE_PREPEND1: //prepends line 6 value to line 1
				case HUDMSGTYPE_PREPEND2: //prepends line 6 counter index value to line 1
				{
					for (uint i = 0; i < TextIDs.length(); i++)
					{
						if (i == 5)
						{
							lines[0] = "" + ((msgType == HUDMSGTYPE_PREPEND2) ? BP::Game::GetCounter(TextIDs[i]) : TextIDs[i]);
						}
						else if (i == 4 && (flags & (HUDMSG_FLAG_SOUND|HUDMSG_FLAG_CUSTOMSOUND)) != 0)
						{
							//do nothing
						}
						else
						{
							if (TextIDs[i] >= 0)
							{
								if (customTextMask & (1 << i) != 0)
									lines.insertLast(Game.GetLocalizedText("$str_" + TextIDs[i]));
								else
									lines.insertLast(BP::Game::GetLocalText(TextIDs[i], language));
							}
							else if (i == 0)
							{
								lines.insertLast("");
							}
						}
					}

					break;
				}
				case HUDMSGTYPE_SHOWKEYSLEFT:
				{
					int nKeys, remainingKeys, keyBits;
					if (!Game.GetHubKeyInfo(hubID, nKeys, remainingKeys, keyBits))
						return;
					
					if (remainingKeys <= 0)
					{
						lines.insertLast(Game.GetLocalizedText("$str_005")); //this level found
						lines.insertLast(Game.GetLocalizedText("$str_004")); //all keys on
					}
					else if (remainingKeys == 1)
					{
						lines.insertLast(Game.GetLocalizedText("$str_003")); //on this level
						lines.insertLast(Game.GetLocalizedText("$str_002")); //1 Key to Find
					}
					else
					{
						lines.insertLast(Game.GetLocalizedText("$str_001")); //on this level
						lines.insertLast("" + remainingKeys + " " + Game.GetLocalizedText("$str_000")); //x Keys to Find
					}
					
					break;
				}
				default:
				{
					for (uint i = 0; i < TextIDs.length(); i++)
					{
						if (i == 4 && (flags & (HUDMSG_FLAG_SOUND|HUDMSG_FLAG_CUSTOMSOUND)) != 0)
						{
							//do nothing
						}
						else if (TextIDs[i] >= 0)
						{
							if (customTextMask & (1 << i) != 0)
								lines.insertLast(Game.GetLocalizedText("$str_" + TextIDs[i]));
							else
								lines.insertLast(BP::Game::GetLocalText(TextIDs[i], language));
						}
					}
					
					break;
				}
			}
			
			if ((flags & (HUDMSG_FLAG_SOUND|HUDMSG_FLAG_CUSTOMSOUND)) != 0)
				PlaySound();
			
			int textTime = 120;
			if (textVisibleTime > 0)
				textTime = textVisibleTime * 60;
			
			if (flags & HUDMSG_FLAG_HELP != 0)
			{
				BP::HUD::ShowHelp(lines, textTime + 60);
			}
			else
			{
				int alignment = BP::HUD::MSG_TOP;
				if (flags & HUDMSG_FLAG_ALIGN_CENTER != 0)
					alignment = BP::HUD::MSG_CENTER;
				BP::HUD::ShowMessage(lines, textTime, alignment);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlaySound()
		{
			int soundID = TextIDs[4];
			if (soundID > 0)
			{
				if (flags & HUDMSG_FLAG_CUSTOMSOUND != 0)
					playerActor.PlaySound(BP::Game::GetCustomSound(soundID));
				else
					playerActor.PlaySoundWithLookupID(soundID);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
