//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_SectorModifier
// Author      : BehemothProgrammer
// Description : Can modify sector area flags, or the 6 args, or water/sector height
// OnTrigger   : Starts modification of sector. If return delay is 0 then sets persistent bit.
//
// Parameters  : (7) Params 1 -128..127 				: Sector Modifier Type
//               (6) Params 2 -128..127 				: AreaFlagType - Area Flag Bit to turn on (1..32 negative values turn that bit off)
//														: AreaArgType - The arg to set. 1..6
//               Radius									: Start Delay
//               WallRadius								: Return delay (0 = no reset)
//               Height 								: Target value (AreaArg(-32768..32767), Sector/Water Height)
//               StepHeight/DeadHeight					: Time to reach target value (AreaArg, Sector/Water Height)
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: (Sound ID lookup. 526 is standard pillar move sound.)
//               (2) Health -32768..32767				: (Target Yaw Offset in degrees)
//               Spawn Flags 1 - Screen Shake			: Will shake the screen when moves
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum SectorModifierTypes
	{
		SM_TYPE_AREAFLAG = 0,
		SM_TYPE_AREAARG = 1,
		SM_TYPE_SECTORHEIGHT = 2, //also moves actor
		SM_TYPE_WATERHEIGHT = 3, //also moves actor
		SM_TYPE_AREATELEPORTARGS = 4, //Changes Area Arg1 and 2. Height is arg1 (TID) and Heath becomes arg2 (MapID). Param 2 - AreaArgType is not used.
		SM_TYPE_ACTORHEIGHT = 5, //only moves actor
	}

	enum SectorModifierFlags
	{
		SM_FLAG_PERPETUAL = (1 << 0), //After return has finished will start moving again. Will also always return even with delay 0
		SM_FLAG_PLAYERGROUNDED = (1 << 1), //Triggers only when the player is grounded
		SM_FLAG_NOACTORMOVE = (1 << 2), //Will not move this actor up/down for Sector or Water Modifier Types
		SM_FLAG_PLAYERROTATE = (1 << 3), //If the player is grounded and on the area then the player will rotate too
		SM_FLAG_CUSTOMSOUND = (1 << 4), //use custom sounds from BP_GameDefCustom actor def
		SM_FLAG_SOUNDEND = (1 << 5), //Heath becomes the soundID to play when finishes movement. To and from.
		SM_FLAG_PERSISTTRIGGER = (1 << 6), //calls trigger onbeginlevel when persistent marked which will include delay times/sounds/shakes.
	}

	enum SectorModifierStates
	{
		SMS_NONE = 0,
		SMS_STARTIDLE = 1,
		SMS_STARTDELAY = 2,
		SMS_STARTING = 3,
		SMS_RETURNIDLE = 4,
		SMS_RETURNDELAY = 5,
		SMS_RETURNING = 6
	}

	class CL_BP_SectorModifier : CL_BP_ScriptObject
	{
		int modifierType;
		int areaFlagBit;
		uint areaFlags;
		int areaArg;
		int targetArgValue;
		int targetArgValue2;
		int flags;
		int moveSoundID;
		float targetYawOffset;
		float startDelay;
		float returnDelay;
		float targetHeight;
		float targetTime;
		
		kVec3 startOrigin;
		int startSector;
		int startArea;
		array<int> startAreaArgs;
		float startActorZ;
		float startSectorHeight;
		float startWaterHeight;
		float startYaw;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_SectorModifier(kActor @actor)
		{
			super(actor);
			state = SMS_NONE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			modifierType = self.SpawnParams(7);
			areaFlagBit = BP::Math::Clamp(self.SpawnParams(6), -32, 32);
			areaFlags = uint(1 << (Math::Abs(areaFlagBit) - 1));
			areaArg = BP::Math::Clamp(self.SpawnParams(6), 1, 6) - 1;
			flags = self.SpawnParams(3);
			moveSoundID = self.SpawnParams(0);
			targetYawOffset = Math::Deg2Rad(self.SpawnParams(2));
			targetArgValue2 = self.SpawnParams(2);
			startDelay = self.Radius();
			returnDelay = self.WallRadius();
			targetHeight = self.Height();
			targetArgValue = int16(self.Height());
			targetTime = self.StepHeight();
			
			startOrigin = self.Origin();
			startArea = self.AreaID();
			startSector = self.SectorIndex();
			for (int i = 0; i < 6; i++)
				startAreaArgs.insertLast(World.GetAreaArg(startArea, i));
			startActorZ = self.Origin().z;
			startYaw = self.Yaw();
			startSectorHeight = self.FloorHeight();
			startWaterHeight = self.GetWaterHeight();
			
			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
			
			if (self.IsPersistentMarked())
			{
				if (flags & SM_FLAG_PERPETUAL != 0)
				{
					if (startDelay > 0.0f)
						SetState(SMS_STARTDELAY);
					else
						SetState(SMS_STARTING);
				}
				else
				{
					if (flags & SM_FLAG_PERSISTTRIGGER != 0)
					{
						SetState(SMS_STARTIDLE);
						OnActivateTrigger();
					}
					else
					{
						SetState(SMS_RETURNIDLE);
					}
				}
			}
			else
			{
				SetState(SMS_STARTIDLE);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			TickState();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			OnActivateTrigger();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivateTrigger()
		{
			self.Flags() &= ~AF_ACTIVATED;
			
			if (flags & SM_FLAG_PLAYERGROUNDED != 0 && !Player.Actor().OnGround())
				return;
			
			if (state != SMS_STARTIDLE)
				return;
			
			if (returnDelay <= 0.0f)
				self.MarkPersistentBit(false);
			
			if (startDelay > 0.0f)
				SetState(SMS_STARTDELAY);
			else
				SetState(SMS_STARTING);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			stateStartTime = BP::Time::FromNow(0.0f);
			switch (state)
			{
				case SMS_STARTIDLE:
				{
					if (prevState != OBJ_STATE_NONE)
						ModifySector(0.0f);
					break;
				}
				case SMS_STARTDELAY:
				{
					ModifySector(0.0f);
					stateEndTime = BP::Time::FromNow(startDelay);
					break;
				}
				case SMS_STARTING:
				{
					PlayMoveSound();
					MoveShake();
					ModifySector(0.0f);
					stateEndTime = BP::Time::FromNow(targetTime);
					break;
				}
				case SMS_RETURNIDLE:
				{
					ModifySector(1.0f);
					break;
				}
				case SMS_RETURNDELAY:
				{
					ModifySector(1.0f);
					stateEndTime = BP::Time::FromNow(returnDelay);
					break;
				}
				case SMS_RETURNING:
				{
					PlayMoveSound();
					MoveShake();
					ModifySector(1.0f);
					stateEndTime = BP::Time::FromNow(targetTime);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TickState()
		{
			switch (state)
			{
				case SMS_STARTIDLE:
				{
					break;
				}
				case SMS_STARTDELAY:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(SMS_STARTING);
					break;
				}
				case SMS_STARTING:
				{
					float normTime = BP::Time::Normalize(stateStartTime, stateEndTime);
					ModifySector(normTime);
					if (normTime >= 1.0f)
					{
						self.StopSound();
						if (flags & SM_FLAG_SOUNDEND != 0)
							PlayEndSound();

						if (returnDelay > 0.0f)
							SetState(SMS_RETURNDELAY);
						else if (flags & SM_FLAG_PERPETUAL != 0)
							SetState(SMS_RETURNING);
						else
							SetState(SMS_RETURNIDLE);
					}
					break;
				}
				case SMS_RETURNIDLE:
				{
					break;
				}
				case SMS_RETURNDELAY:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(SMS_RETURNING);
					break;
				}
				case SMS_RETURNING:
				{
					float normTime = BP::Time::Normalize(stateStartTime, stateEndTime);
					ModifySector(1.0f - normTime);
					if (normTime >= 1.0f)
					{
						self.StopSound();
						if (flags & SM_FLAG_SOUNDEND != 0)
							PlayEndSound();
						
						if (flags & SM_FLAG_PERPETUAL != 0)
						{
							if (startDelay > 0.0f)
								SetState(SMS_STARTDELAY);
							else
								SetState(SMS_STARTING);
						}
						else
						{
							SetState(SMS_STARTIDLE);
						}
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ModifySector(const float normTime)
		{
			if (targetYawOffset != 0.0f && modifierType != SM_TYPE_AREATELEPORTARGS && (flags & SM_FLAG_SOUNDEND) == 0)
			{
				kAngle lastYaw = self.Yaw();
				self.Yaw() = BP::Math::SmoothStepFloat(startYaw, startYaw + targetYawOffset, normTime); 
				if (flags & SM_FLAG_PLAYERROTATE != 0)
				{
					kActor@ pActor = Player.Actor().CastToActor();
					if (pActor.AreaID() == startArea && pActor.OnGround())
						pActor.Yaw() += self.Yaw().Diff(lastYaw);
				}
			}
			
			switch (modifierType)
			{
				case SM_TYPE_AREAFLAG:
				{
					if (areaFlagBit != 0 && (normTime >= 1.0f || normTime <= 0.0f))
					{
						bool on = normTime >= 1.0f;
						if (areaFlagBit < 0)
							on = !on;
						World.FloodMatchingAreaFlags(startSector, areaFlags, on);
					}
					break;
				}
				case SM_TYPE_AREAARG:
				{
					World.ChangeAreaArg(startArea, areaArg, BP::Math::LerpInt(startAreaArgs[areaArg], targetArgValue, normTime));
					break;
				}
				case SM_TYPE_SECTORHEIGHT:
				{
					if (flags & SM_FLAG_NOACTORMOVE == 0)
						self.Origin().z = BP::Math::SmoothStepFloat(startActorZ, startActorZ + (targetHeight - startSectorHeight), normTime);
					World.ChangeSectorHeight(startSector, BP::Math::SmoothStepFloat(startSectorHeight, targetHeight, normTime));
					//Log("SM TID: " + self.TID() + " FloorHeight: " + self.FloorHeight());
					break;
				}
				case SM_TYPE_WATERHEIGHT:
				{
					if (flags & SM_FLAG_NOACTORMOVE == 0)
						self.Origin().z = BP::Math::SmoothStepFloat(startActorZ, startActorZ + (targetHeight - startWaterHeight), normTime);
					World.ChangeAreaWaterHeight(startArea, BP::Math::SmoothStepFloat(startWaterHeight, targetHeight, normTime));
					break;
				}
				case SM_TYPE_AREATELEPORTARGS:
				{
					World.ChangeAreaArg(startArea, AREAARG_WARP_TID, BP::Math::LerpInt(startAreaArgs[AREAARG_WARP_TID], targetArgValue, normTime));
					World.ChangeAreaArg(startArea, AREAARG_WARP_MAPID, BP::Math::LerpInt(startAreaArgs[AREAARG_WARP_MAPID], targetArgValue2, normTime));
					break;
				}
				case SM_TYPE_ACTORHEIGHT:
				{
					self.Origin().z = BP::Math::SmoothStepFloat(startActorZ, targetHeight, normTime);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlayMoveSound()
		{
			if (moveSoundID > 0)
			{
				if (flags & SM_FLAG_CUSTOMSOUND != 0)
					self.PlaySound(BP::Game::GetCustomSound(moveSoundID));
				else
					self.PlaySoundWithLookupID(moveSoundID);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlayEndSound()
		{
			if (targetArgValue2 > 0)
			{
				if (flags & SM_FLAG_CUSTOMSOUND != 0)
					self.PlaySound(BP::Game::GetCustomSound(targetArgValue2));
				else
					self.PlaySoundWithLookupID(targetArgValue2);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MoveShake()
		{
			//has screen shake spawn flag on
			if ((self.SpawnFlags1() & (1 << 30)) != 0)
				BP::Game::AddShake(self, targetTime);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
