//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_ShootTrigger
// Author      : BehemothProgrammer
// Description : When shot will trigger actors/script. 
// OnTrigger   : Trigger actors/script and mark persistent.
//
// Parameters  : (7) Params 1 -128..127					: Impact Type of this shootable trigger (common: 2=metal,4=human,5=creature)
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: Actors TID to trigger. If is less than 1 this will not be used.
//               StepHeight/DeadHeight					: Sound ID to play when triggered
//               WallRadius								: Amount of time to wait until can be triggered again (If is repeatable)
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum ShootTriggerFlags
	{
		ST_FLAG_REPEATABLE = (1 << 0), //(1) can be triggered multiple times
		ST_FLAG_TRIGGER_ONDEATH = (1 << 1), //(2) triggers OnDeath instead of OnDamage
		ST_FLAG_FLASH_ONDAMAGE = (1 << 2), //(4) flash when shot
		ST_FLAG_EXPLOSIVE = (1 << 3), //(8) only explosives
		ST_FLAG_NONEXPLOSIVE = (1 << 4), //(16) only non explosives
		ST_FLAG_CUSTOMSOUND = (1 << 5), //(32) use custom sound
		ST_FLAG_STARTNONSOLID = (1 << 6), //(64) starts non solid OnBeginLevel is if not marked persistent
		ST_FLAG_ONPLAYER = (1 << 7), //(128) plays sound on playerActor
	}

	enum ShootTriggerStates
	{
		STS_IDLE = 0,
		STS_TRIGGERED,
		STS_IDLETRIGGERED
	}

	class CL_BP_ShootTrigger : CL_BP_ScriptObject
	{
		int flags;
		int impact;
		int startHealth;
		float startRadius;
		float startHeight;
		int triggerTID;
		int soundID;
		float retriggerDelay;

		int health;
		float retriggerTime;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_ShootTrigger(kActor @actor)
		{
			super(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			//make dead so OnDeath will not get called (for map actors)
			health = self.Health();
			self.Health() = 0;

			flags = self.SpawnParams(3);
			impact = self.SpawnParams(7);
			startHealth = self.SpawnParams(2);
			startRadius = self.Radius();
			startHeight = self.Height();
			triggerTID = self.SpawnParams(0);
			soundID = int(self.StepHeight());
			retriggerDelay = self.WallRadius();

			self.ImpactType() = impact;

			if (self.IsPersistentMarked())
			{
				SetState(STS_IDLETRIGGERED);
			}
			else
			{
				SetState(STS_IDLE);
				if (flags & ST_FLAG_STARTNONSOLID != 0)
					self.Flags() &= ~AF_SOLID;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			DoTrigger();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnDamage(kActor @instigator, kDictMem @damageDef, const int damage)
		{
			if (CanBeDamaged(instigator, damageDef))
			{
				if (flags & ST_FLAG_FLASH_ONDAMAGE != 0)
					self.RunFxEvent("Mantis_Hit");
				
				health -= damage;
				if (health <= 0 || flags & ST_FLAG_TRIGGER_ONDEATH == 0)
				{
					DoTrigger();
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool CanBeDamaged(kActor@ instigator, kDictMem @damageDef)
		{
			bool explosiveDamage = false;
			if (damageDef !is null)
				damageDef.GetBool(DEF_DAMAGE_EXPLODE, explosiveDamage);
			
			if (flags & ST_FLAG_EXPLOSIVE != 0)
				return explosiveDamage;
			else if (flags & ST_FLAG_NONEXPLOSIVE != 0)
				return !explosiveDamage;
			
			return true;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoTrigger()
		{
			if (BP::Time::Reached(retriggerTime))
			{
				retriggerTime = BP::Time::FromNow(retriggerDelay);
				SetState(STS_TRIGGERED);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			switch (state)
			{
				case STS_IDLE:
				{
					self.Flags() &= ~(AF_DEAD | AF_INVINCIBLE | AF_ACTIVATED);
					self.Flags() |= AF_SOLID;
					health = startHealth;
					self.Radius() = startRadius;
					self.Height() = startHeight;
					self.MarkPersistentBit(true);
					break;
				}
				case STS_TRIGGERED:
				{
					PlaySound();
					
					if (triggerTID > 0)
						World.TriggerActorsByTID(self, triggerTID);

					//if repeatable go back to idle state
					if (flags & ST_FLAG_REPEATABLE != 0)
						SetState(STS_IDLE);
					else
						SetState(STS_IDLETRIGGERED);
					
					break;
				}
				case STS_IDLETRIGGERED:
				{
					self.Flags() |= AF_DEAD | AF_INVINCIBLE | AF_ACTIVATED;
					self.Flags() &= ~AF_SOLID;
					self.Radius() = 0.0f;
					self.Height() = 0.0f;
					self.MarkPersistentBit(false);
					
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlaySound()
		{
			if (soundID > 0)
			{
				kActor@ soundActor = (flags & ST_FLAG_ONPLAYER != 0) ? playerActor : self;
				if (flags & ST_FLAG_CUSTOMSOUND != 0)
					soundActor.PlaySound(BP::Game::GetCustomSound(soundID));
				else
					soundActor.PlaySoundWithLookupID(soundID);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
