//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_Destructible
// Author      : BehemothProgrammer
// Description : Once this actor is killed it will play it's destructibleDeath animation and
//               can trigger other actors. And can drop an item with a percent chance.
// OnTrigger   : Nothing
//
// Parameters  : (3) Max Regenerations -32768..32767	: Flags
//               WallRadius								: ActorID to spawn when destroyed (Item Drop) If is less than 1 this will not be used.
//               (6) Params 2 -128..127					: 0..100% chance of spawning the item drop.
//               (0) Target TID -32768..32767 			: Actors TID to trigger when destroyed. If is less than 1 this will not be used.
//               StepHeight/DeadHeight					: Fx to spawn when destroyed

// Extra Defs  : DestructibleFlags						: Sets flags in .ctor (for spawned actors)
//               DestructibleItemChance					: Sets itemChance in .ctor (for spawned actors)
//               DestructibleFX							: Sets FxID in .ctor (for spawned actors)
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum DestructibleFlags
	{
		DESTRUCTIBLE_FLAG_ONLY_EXPLOSION_IMPACTS = (1 << 0),
		DESTRUCTIBLE_FLAG_ONLY_NONEXPLOSION_IMPACTS = (1 << 1),
		DESTRUCTIBLE_FLAG_REMOVE_ONDEATH = (1 << 2),
		DESTRUCTIBLE_FLAG_FXCUSTOM = (1 << 3), //use custom fx from BP_GameDefCustom actor def
		DESTRUCTIBLE_FLAG_FLASH_ONDAMAGE = (1 << 4),
		DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP = (1 << 5)
	}

	class CL_BP_Destructible : CL_BP_ScriptObject
	{
		int flags;
		int itemChance;
		int triggerTID;
		int health;
		int fxID;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_Destructible(kActor @actor)
		{
			super(actor);
			itemChance = 100;
			flags = BP::Actor::GetDefInt(self, DEF_DESTRUCTIBLE_FLAGS);
			itemChance = BP::Actor::GetDefInt(self, DEF_DESTRUCTIBLE_ITEMCHANCE, itemChance);
			fxID = BP::Actor::GetDefInt(self, DEF_DESTRUCTIBLE_FX);
			
			//make dead so OnDeath will not get called (for spawned actors)
			health = self.Health();
			self.Health() = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			flags = self.SpawnParams(3);
			triggerTID = self.SpawnParams(0);
			itemChance = BP::Math::Clamp(self.SpawnParams(6), 0, 100);
			fxID = int(self.StepHeight());
			
			//make dead so OnDeath will not get called (for map actors)
			health = self.Health();
			self.Health() = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			if (flags & DESTRUCTIBLE_FLAG_REMOVE_ONDEATH != 0)
			{
				self.Remove();
				return;
			}
			
			self.Flags() |= AF_DEAD | AF_IGNORESOUNDEVENTS | AF_INVINCIBLE;
			self.Flags() &= ~AF_SOLID;
			self.Radius() = 0.0f;
			self.Height() = 0.0f;
			if (self.AnimState().CheckAnimID(anim_destructibleDeath))
			{
				self.AnimState().Set(anim_destructibleDeath, 4.0f, 0);
				self.AnimState().SetLastFrame(true);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnSleep()
		{
			if (flags & DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP != 0)
				self.Flags() |= AF_NODRAW;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnWake()
		{
			if (flags & DESTRUCTIBLE_FLAG_NODRAW_ONSLEEP != 0)
				self.Flags() &= ~AF_NODRAW;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnDamage(kActor @instigator, kDictMem @damageDef, const int damage)
		{
			if (CanBeDamaged(instigator, damageDef))
			{
				health -= damage;
				if (health > 0)
				{
					if (flags & DESTRUCTIBLE_FLAG_FLASH_ONDAMAGE != 0)
						self.RunFxEvent("Mantis_Hit");
				}
				else
				{
					//Do Death
					self.Flags() |= AF_DEAD | AF_INVINCIBLE;
					self.Flags() &= ~AF_SOLID;
					self.Radius() = 0.0f;
					self.Height() = 0.0f;
					if (self.AnimState().CheckAnimID(anim_destructibleDeath))
						self.AnimState().Blend(anim_destructibleDeath, 4.0f, 4.0f, 0);
					self.MarkPersistentBit(false);
					
					//Spawn Fx
					if (flags & DESTRUCTIBLE_FLAG_FXCUSTOM != 0)
						self.SpawnFx(BP::Game::GetCustomFx(fxID), Math::vecZero, self.Rotation());
					else
						self.SpawnFx(BP::Game::GetFx(fxID), Math::vecZero, self.Rotation());
					
					//Trigger Actors
					if (triggerTID > 0)
						World.TriggerActorsByTID(self, triggerTID);
					
					//Try spawning item drop
					int actorID = int(self.WallRadius());
					if (actorID > 0 && Math::RandMax(100) < itemChance)
					{
						kActor@ item = BP::Actor::Spawn(actorID, self.Origin(), self.Yaw(), self.SectorIndex());
						if (BP::Actor::Exists(item))
							BP::Actor::TossItem(item);
					}
					
					if (flags & DESTRUCTIBLE_FLAG_REMOVE_ONDEATH != 0)
						self.Remove();
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		bool CanBeDamaged(kActor@ instigator, kDictMem @damageDef)
		{
			bool explosiveDamage = false;
			if (damageDef !is null)
				damageDef.GetBool(DEF_DAMAGE_EXPLODE, explosiveDamage);
			
			if (flags & DESTRUCTIBLE_FLAG_ONLY_EXPLOSION_IMPACTS != 0)
				return explosiveDamage;
			else if (flags & DESTRUCTIBLE_FLAG_ONLY_NONEXPLOSION_IMPACTS != 0)
				return !explosiveDamage;
			
			return true;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MeleeBluntHeavy(kActor @instigator, const float w, const float x, const float y, const float z)
		{
			self.MeleeObject("Damage_Blunt_20", kVec3(x, y, z), w);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
