//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_Door
// Author      : BehemothProgrammer
// Description : Plays the door animations with delays.

//				Use key frame action event 23 (Unused23) to play custom sounds by ID (arg1). Define your custom sounds in your BP_GameDefCustom actor def.
//				Use key frame action event 24 (Unused24) to spawn fx by ID (arg1).
//				Use key frame action event 60 (Unused60) to spawn custom fx by ID (arg1). Define your custom fx in your BP_GameDefCustom actor def.
//
//				---Animation IDs---
//				anim_doorOpen  - Opening
//				anim_doorClose - Closing
//				anim_doorIdle  - Idle Closed (optional)
//				anim_trapIdle  - Idle Opened (optional)
//
// OnTrigger   : Activates the door to open to attempt to open/close
//
// Parameters  : (3) Max Regenerations -32768..32767	: Flags
//               WallRadius								: Open delay (0 = no auto open after opening if BP_DF_FLAG_STARTOPEN is ON)
//               StepHeight/DeadHeight					: Close delay (0 = no auto close after opening if BP_DF_FLAG_STARTOPEN is OFF)
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum DoorFlags
	{
		DF_FLAG_STARTOPEN = (1 << 0), //Starts in open state instead of closed
	}

	enum DoorAnimSoundEventFlags
	{
		DASEF_STOP = (1 << 0), //calls StopSound on this actor
	}

	enum DoorStates
	{
		DS_OPENED = 0,
		DS_CLOSED = 1,
		DS_OPENING = 2,
		DS_CLOSING = 3,
		DS_OPENDELAY = 4,
		DS_CLOSEDELAY = 5
	}

	class CL_BP_Door : CL_BP_ScriptObject
	{
		float openDelay;
		float closeDelay;
		int flags;
		int health;
		
		bool hasIdleCloseAnim;
		bool hasIdleOpenAnim;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_Door(kActor @actor)
		{
			super(actor);
			state = DS_CLOSED;
			
			//make dead so OnDeath will not get called (for spawned actors)
			health = self.Health();
			self.Health() = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			//make dead so OnDeath will not get called (for map actors)
			health = self.Health();
			self.Health() = 0;
			
			hasIdleCloseAnim = self.AnimState().CheckAnimID(anim_doorIdle);
			hasIdleOpenAnim = self.AnimState().CheckAnimID(anim_trapIdle);

			openDelay = self.WallRadius();
			closeDelay = self.StepHeight();
			flags = self.SpawnParams(3);
			
			self.Flags() |= AF_IGNORESOUNDEVENTS | AF_INVINCIBLE;
			
			if (self.IsPersistentMarked())
			{
				self.Flags() |= AF_ACTIVATED;
				if (flags & DF_FLAG_STARTOPEN != 0)
				{
					SetState(DS_CLOSED);
				}
				else
				{
					SetState(DS_OPENED);
				}
			}
			else
			{
				if (flags & DF_FLAG_STARTOPEN != 0)
				{
					SetState(DS_OPENED);
				}
				else
				{
					SetState(DS_CLOSED);
				}
			}
			
			self.AnimState().SetLastFrame(true);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			TickState();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoTrigger()
		{
			self.Flags() &= ~AF_IGNORESOUNDEVENTS;
			
			if (flags & DF_FLAG_STARTOPEN != 0)
			{
				if (state != DS_OPENED)
					return;
				
				//starts open
				if (openDelay <= 0.0f)
				{
					self.Flags() |= AF_ACTIVATED;
					self.MarkPersistentBit(false);
				}
				
				SetState(closeDelay > 0.0f ? DS_CLOSEDELAY : DS_CLOSING);
			}
			else
			{
				if (state != DS_CLOSED)
					return;
				
				//starts closed
				if (closeDelay <= 0.0f)
				{
					self.Flags() |= AF_ACTIVATED;
					self.MarkPersistentBit(false);
				}
				
				SetState(openDelay > 0.0f ? DS_OPENDELAY : DS_OPENING);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			self.Flags() &= ~AF_ACTIVATED;
			DoTrigger();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlayAnim(const int animID)
		{
			if (self.AnimState().PlayingID() == animID)
				return;
			
			if (self.GameTicks() > 0)
			{
				self.AnimState().Blend(animID, 4.0f, 4.0f, 0);
			}
			else
			{
				self.AnimState().Set(animID, 4.0f, 0);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnStateChange() override
		{
			stateStartTime = BP::Time::FromNow(0.0f);
			switch (state)
			{
				case DS_OPENDELAY:
				{
					stateEndTime = BP::Time::FromNow(openDelay);
					break;
				}
				case DS_OPENING:
				{
					PlayAnim(anim_doorOpen);
					break;
				}
				case DS_OPENED:
				{
					PlayAnim(anim_doorOpen);
					break;
				}
				case DS_CLOSEDELAY:
				{
					stateEndTime = BP::Time::FromNow(closeDelay);
					break;
				}
				case DS_CLOSING:
				{
					PlayAnim(anim_doorClose);
					break;
				}
				case DS_CLOSED:
				{
					PlayAnim(anim_doorClose);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TickState()
		{
			switch (state)
			{
				case DS_OPENDELAY:
				{
					if (BP::Time::Reached(stateEndTime))
						SetState(DS_OPENING);
					break;
				}
				case DS_OPENING:
				{
					//Log("TickState DS_OPENING cycleCompleted: " + self.AnimState().CycleCompleted() + " trackTime: " + self.AnimState().TrackTime());
					if (self.AnimState().CycleCompleted())
						SetState(DS_OPENED);
					break;
				}
				case DS_OPENED:
				{
					//open idle loop
					if (self.AnimState().PlayingID() == anim_doorOpen && self.AnimState().CycleCompleted() && hasIdleOpenAnim)
					{
						self.Flags() &= ~AF_IGNORESOUNDEVENTS;
						PlayAnim(anim_trapIdle);
					}

					// if (self.GameTicks() % 60 == 0)
						// Log("TickState DS_OPENED startClose: " + (flags & DF_FLAG_STARTOPEN == 0) +
						// " closeDelay: " + closeDelay +
						// " TimeReached: " + BP::Time::Reached(stateEndTime) + " stateEndTime: " + stateEndTime);
					//if should automatically go back to close delay
					if (flags & DF_FLAG_STARTOPEN == 0 && closeDelay > 0.0f)// && BP::Time::Reached(stateEndTime))
					{
						SetState(DS_CLOSEDELAY);
					}
					break;
				}
				case DS_CLOSEDELAY:
				{
					//Log("TickState DS_CLOSEDELAY timeReached: " + BP::Time::Reached(stateEndTime) + " stateEndTime: " + stateEndTime + " curTime: " + BP::Time::GameTime());
					if (BP::Time::Reached(stateEndTime))
						SetState(DS_CLOSING);
					break;
				}
				case DS_CLOSING:
				{
					//Log("TickState DS_CLOSING cycleCompleted: " + self.AnimState().CycleCompleted() + " trackTime: " + self.AnimState().TrackTime());
					if (self.AnimState().CycleCompleted())
						SetState(DS_CLOSED);
					break;
				}
				case DS_CLOSED:
				{
					//close idle loop
					if (self.AnimState().PlayingID() == anim_doorClose && self.AnimState().CycleCompleted() && hasIdleCloseAnim)
					{
						self.Flags() &= ~AF_IGNORESOUNDEVENTS;
						PlayAnim(anim_doorIdle);
					}

					// if (self.GameTicks() % 60 == 0)
						// Log("TickState DS_CLOSED startClose: " + (flags & DF_FLAG_STARTOPEN == 0) +
						// " openDelay: " + openDelay +
						// " TimeReached: " + BP::Time::Reached(stateEndTime) + " stateEndTime: " + stateEndTime);
					//if should automatically go back to open delay
					if (flags & DF_FLAG_STARTOPEN != 0 && openDelay > 0.0f)// && BP::Time::Reached(stateEndTime))
					{
						SetState(DS_OPENDELAY);
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Play custom sound event by SoundID value
		//------------------------------------------------------------------------------------------------------------------------
		void Unused23(kActor @instigator, const float arg1, const float arg2, const float arg3, const float arg4)
		{
			if ((self.Flags() & AF_IGNORESOUNDEVENTS) != 0)
				return;
			
			int soundID = int(arg1);
			int soundFlags = int(arg2);
			if (soundFlags & DASEF_STOP != 0)
			{
				self.StopSound();
			}
			else
			{
				if (BP::Game::HasCustomSound(soundID))
					self.PlaySound(BP::Game::GetCustomSound(soundID));
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Play fx
		//------------------------------------------------------------------------------------------------------------------------
		void Unused24(kActor @instigator, const float arg1, const float x, const float y, const float z)
		{
			if ((self.Flags() & AF_IGNORESOUNDEVENTS) != 0)
				return;
			
			int fxID = int(arg1);
			if (BP::Game::HasFx(fxID))
				self.SpawnFx(BP::Game::GetFx(fxID), kVec3(x, y, z), self.Rotation());
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Play custom fx from BP_GameDefCustom actor def
		//------------------------------------------------------------------------------------------------------------------------
		void Unused60(kActor @instigator, const float arg1, const float x, const float y, const float z)
		{
			if ((self.Flags() & AF_IGNORESOUNDEVENTS) != 0)
				return;
			
			int fxID = int(arg1);
			if (BP::Game::HasCustomFx(fxID))
				self.SpawnFx(BP::Game::GetCustomFx(fxID), kVec3(x, y, z), self.Rotation());
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
