//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_KeyPanel
// Author      : BehemothProgrammer
// Description : A place to put keys to trigger other actors.
// OnTrigger   : If has the key it will activate the hubPanelPiece.
//				 Once all keys have been placed, triggering this actor will do nothing and actors/script can be triggered.
//
// Parameters  : (7) Params 1 -128..127					: Key Bit Offset
//               (6) Params 2 -128..127					: Key Texture
//               (3) Max Regenerations -32768..32767	: Flags
//               (4) Attack Chance 0..255				: HUB ID Key Check 1..7 (normally the first keys are for hub 2)
//               (0) Target TID -32768..32767			: TID to trigger when activated if is greater than 0
//               (2) Health -32768..32767				: Script to trigger when activated if is greater than 0
//               StepHeight/DeadHeight					: Amount of keys to check for 1..32
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum KeyPanelFlags
	{
		KPF_NOSOUND = (1 << 0) //Don't play sound when placing keys
	}

	const int KEYPANEL_STARTNODE = 1;

	class CL_BP_KeyPanel : CL_BP_ScriptObject
	{
		int flags;
		int keyBitOffset;
		int keyCount;
		int keyTexture;
		int hubID;
		int triggerTID;
		int triggerScriptID;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_KeyPanel(kActor @actor)
		{
			super(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			flags = self.SpawnParams(3);
			keyBitOffset = self.SpawnParams(7);
			keyCount = BP::Math::Clamp(int(self.StepHeight()), 1, 32);
			keyTexture = self.SpawnParams(6);
			hubID = self.SpawnParams(4);
			triggerTID = self.SpawnParams(0);
			triggerScriptID = self.SpawnParams(2);

			kStr gvKey = GV_KEY_PANELPIECES;
			int pieces = BP::GameVar::GetInt(gvKey + (hubID - 1));
			for (int i = 0; i < keyCount; i++)
			{
				int keyBit = keyBitOffset + i;
				if (keyBit > 31)
					break;

				bool hasPiece = pieces & (1 << keyBit) != 0;
				self.RenderModel().SetVisibility(KEYPANEL_STARTNODE + i, hasPiece);
				self.RenderModel().SetTexture(KEYPANEL_STARTNODE + i, keyTexture);
			}

			self.Flags() |= AF_INVINCIBLE;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnRestore()
		{
			self.Flags() |= AF_ACTIVATED;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			int nKeys, remainingKeys, keyBits;
			if (!Game.GetHubKeyInfo(hubID, nKeys, remainingKeys, keyBits))
				return;

			//get pieces plaved and set model node visibility
			kStr gvKey = GV_KEY_PANELPIECES;
			gvKey += "" + (hubID - 1);
			int pieces = BP::GameVar::GetInt(gvKey);
			int lastPieceCount = BP::Math::BitCount(pieces, keyBitOffset, keyCount);
			int pieceCount = 0;
			for (int i = 0; i < keyCount; i++)
			{
				int keyBit = keyBitOffset + i;
				if (keyBit > 31)
					break;

				bool hasKey = ((keyBits & (1 << keyBit)) != 0);
				if (hasKey)
				{
					pieces |= (1 << keyBit);
					pieceCount++;
				}
				else
				{
					pieces &= ~(1 << keyBit);
				}
				self.RenderModel().SetVisibility(KEYPANEL_STARTNODE + i, hasKey);
			}

			//1 or more pieces were placed
			if (pieceCount != lastPieceCount)
			{
				BP::GameVar::SetInt(gvKey, pieces);
				if (flags & KPF_NOSOUND == 0)
					self.PlaySound("sounds/shaders/health_pickup_1.ksnd");
			}

			//Has all pieces
			if (pieceCount >= keyCount)
			{
				self.MarkPersistentBit(false);

				//trigger other actors/script
				if (triggerTID > 0)
					World.TriggerActorsByTID(self, triggerTID);
				if (triggerScriptID > 0)
					Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
			}
			else
			{
				self.Flags() &= ~AF_ACTIVATED;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
