//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_Map
// Author      : BehemothProgrammer
// Description : Map manager actor that should be spawned in via script only and never placed in the map itself with the editor.
// OnTrigger   : Nothing.
//
// Parameters  : 
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	int mapActorCount;

	class CL_BP_Map : CL_BP_ScriptObject
	{
		float playerKnifeAttackTime;
		
		kActor@ musicSource; //The musics actor audio source
		int musicLastFlags; //the previous frames music flags
		array<kStr> musicTracks; //list of tracks sound shaders
		array<float> musicTracksDuration; //list of tracks durations
		int musicMapID; //the mapID the music can play on
		int musicPlayOrder; //current track play order
		int musicTrackIndex; //current track index playing
		float musicTrackTime; //amount of time has been playing the track
		kStr musicTrackPath; //current track sound shader path
		array<int> musicAvailTrackIndexs; //list of track indexes available for unique random play order 
		int musicTracksPlayed; //number of tracks played
		float hudHelpTime;
		kStr hudHelpMsg;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_Map(kActor @actor)
		{
			super(actor);
			if (mapActorCount > 0)
			{
				self.Remove();
				return;
			}
			
			mapActorCount++;
			MapStartSetup();
			int musicFlags = BP::GameVar::GetInt(GV_KEY_MUSICFLAGS);
			if (musicFlags & BP::Music::MF_STOPONNEXTMAP != 0)
			{
				BP::GameVar::SetIntFlag(GV_KEY_MUSICFLAGS, BP::Music::MF_STOPONNEXTMAP, false);
				BP::Music::Stop();
			}
			BP::GameVar::SetIntFlag(GV_KEY_MUSICFLAGS, BP::Music::MF_LISTUPDATE);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			BP::GameVar::SetFloat(GV_KEY_GAMETIME, BP::Time::GameTime() + GAME_DELTA_TIME);
			MusicTick();
			BP::Input::Tick();
			HUDTick();
			
			if (BP::Input::AttackPress() && Player.CurrentWeapon() == TW_WEAPON_KNIFE && BP::Time::Reached(playerKnifeAttackTime) &&
				!Player.Locked() && !BP::GameVar::GetIntFlag(GV_KEY_FLAGS, SF_NOKNIFEINTERACTION))
			{
				playerKnifeAttackTime = BP::Time::FromNow(0.7f);
				self.InteractActorsAtPosition(BP::Player::MeleePos(), FUNC_PLAYERKNIFEATTACK);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnEndLevel()
		{
			mapActorCount = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void HUDTick()
		{
			//Update Help Message
			int gvHelpTime = BP::GameVar::GetInt(GV_KEY_HUDHELPTIME);
			if (gvHelpTime > 0)
			{
				BP::GameVar::SetInt(GV_KEY_HUDHELPTIME, 0);
				hudHelpMsg = BP::GameVar::GetString(GV_KEY_HUDHELP);
				hudHelpTime = BP::Time::FromNow(gvHelpTime / 60.0f);
				
				//check for \n characters and convert them to char 10 (new line) which kStr does automatically by converting "\n" to char 10
				array<kStr> msgLines = BP::String::Split(hudHelpMsg, "\\n");
				int lineCount = int(msgLines.length());
				if (lineCount > 1)
				{
					hudHelpMsg = "";
					for (int i = 0; i < lineCount - 1; i++)
						hudHelpMsg += msgLines[i] + "\n";
					hudHelpMsg += msgLines[lineCount - 1];
				}
			}
			
			if (hudHelpTime > 0.0f)
			{
				Game.PrintHelp(hudHelpMsg);
				if (BP::Time::Reached(hudHelpTime))
				{
					hudHelpTime = 0.0f;
					hudHelpMsg = "";
					Game.PrintHelp("");
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MusicTick()
		{
			int musicFlags = BP::GameVar::GetInt(GV_KEY_MUSICFLAGS);
			//On flags changed
			if (musicLastFlags != musicFlags)
			{
				musicLastFlags = musicFlags;
				if ((musicFlags & BP::Music::MF_LISTUPDATE) != 0)
				{
					BP::GameVar::SetIntFlag(GV_KEY_MUSICFLAGS, BP::Music::MF_LISTUPDATE, false);
					
					bool gotEventData = BP::Music::GetEventData(musicTracks, musicTracksDuration, musicPlayOrder, musicMapID);
					if (gotEventData)
					{
						musicTracksPlayed = 0;
						musicAvailTrackIndexs.resize(0);
						switch (musicPlayOrder)
						{
							case BP::Music::MUSIC_PO_RANDOM:
							{
								MusicSetTrack(BP::Array::GetUniqueElement(musicTracks.length(), musicAvailTrackIndexs));
								break;
							}
							default:
							{
								MusicSetTrack(0);
								break;
							}
						}
					}
				}
			}

			float musicVolume = BP::CVar::GetFloat(CVAR_MUSICVOLUME);
			bool canPlay = !((musicFlags & BP::Music::MF_ISPLAYING) == 0 || (musicFlags & BP::Music::MF_DISABLE) != 0 || BP::String::IsEmpty(musicTrackPath) || 
							(musicMapID != -2 && musicMapID != Game.GetCurrentMapID()) || (BP::Cinema::IsActive() && BP::Actor::IsDead(playerActor)) || musicVolume < 0.1f);
			if (canPlay)
			{
				if (!BP::Actor::Exists(musicSource))
				{
					@musicSource = BP::Actor::Spawn(ACTOR_NAME_MUSIC);
				}
				else
				{
					musicSource.Origin() = (BP::Cinema::IsActive() ? Camera.origin : playerActor.Origin());
					musicSource.PlaySound(musicTrackPath);
				}
				
				float lastTrackTime = musicTrackTime;
				musicTrackTime += GAME_DELTA_TIME;
				if (musicTracksDuration[musicTrackIndex] > 0.0f && musicTrackTime >= musicTracksDuration[musicTrackIndex])
				{
					musicTracksPlayed++;
					switch (musicPlayOrder)
					{
						case BP::Music::MUSIC_PO_RANDOM:
						case BP::Music::MUSIC_PO_RANDOMAFTERFIRST:
						{
							MusicSetTrack(BP::Array::GetUniqueElement(musicTracks.length(), musicAvailTrackIndexs, musicTrackIndex));
							break;
						}
						default:
						{
							MusicSetTrack(BP::Array::NextIndex(musicTracks.length(), musicTrackIndex));
							break;
						}
					}
				}
			}
			else
			{
				musicTrackTime = 0.0f;
				if (BP::Actor::Exists(musicSource))
				{
					musicSource.StopLoopingSounds();
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MusicSetTrack(const int index)
		{
			musicTrackIndex = index;
			musicTrackPath = musicTracks[musicTrackIndex];
			musicTrackTime = 0.0f;
			if (BP::Actor::Exists(musicSource))
			{
				musicSource.StopLoopingSounds();
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PlayerKnifeAttack(kActor @actor, const float arg1, const float arg2, const float arg3, const float arg4)
		{
			if (!BP::Actor::Exists(actor) || BP::Actor::IsDead(actor) || actor.ImpactType() == IT_FORCEFIELD ||
				(actor.Flags() & AF_HIDDEN) != 0 || (actor.Flags() & AF_HOLDTRIGGERANIM) != 0 ||
				(actor.Flags() & AF_INVINCIBLE) != 0)
				return;
			
			int knifeImpactType = BP::Actor::GetDefInt(actor, DEF_KNIFEIMPACT, -1);
			if (knifeImpactType >= 0)
			{
				int frontDamage = BP::Actor::GetDefInt(actor, DEF_KNIFEDAMAGE, 0);
				int backDamage = BP::Actor::GetDefInt(actor, DEF_KNIFEBACKDAMAGE, 0);
				bool noSpawnFx = BP::Actor::GetDefBool(actor, DEF_KNIFENOSPAWNFX, false);
				
				float radius = BP::Player::MELEE_RADIUS;
				kVec3 meleePos = BP::Player::MeleePos();
				float dist = actor.Origin().Distance(meleePos);
				int angle1 = int(Math::Rad2Deg(actor.Yaw())) % 360;
				int angle2 = int(Math::Rad2Deg(playerActor.Yaw())) % 360;
				if(angle1 < 0) angle1 += 360;
				if(angle2 < 0) angle2 += 360;
				int angDiff = (angle1 - angle2);
				bool inRange = false;
				int damage = 0;
				// check for backstabs
				if (dist <= ((radius * 0.5f) + actor.Radius()) && Math::Abs(angDiff) <= 40)
				{
					damage = backDamage;
				}
				// do regular direct damage
				else if (dist <= (radius + actor.Radius()))
				{
					damage = frontDamage;
				}
				
				if (damage <= 0)
					return;
				
				kVec3 impactPos = kVec3(0.0f, 0.0f, meleePos.z);
				kStr fxName;
				kVec3 fxPos;
				switch (knifeImpactType)
				{
					case IT_METAL:
						actor.PlaySound("sounds/shaders/tomahawk_impact_flesh.ksnd");
						fxPos = kVec3(0.0f, 0.0f, meleePos.z);
						fxName = "fx/generic_90.kfx"; //sparks
						break;
					case IT_FLESH_HUMAN:
						actor.PlaySound("sounds/shaders/tomahawk_impact_flesh.ksnd");
						fxPos = impactPos;
						fxName = "fx/generic_244.kfx"; //5 red bloods spit out
						break;
					case IT_FLESH_CREATURE:
						actor.PlaySound("sounds/shaders/tomahawk_impact_flesh.ksnd");
						fxPos = impactPos;
						fxName = "fx/greenblood.kfx";
						break;
					case IT_WOOD:
						actor.PlaySound("sounds/shaders/tomahawk_impact_flesh.ksnd");
						fxPos = impactPos;
						fxName = "BP_Addons/fx/impact_wood.kfx";
						break;
				}
				
				if (!noSpawnFx)
					actor.SpawnFx(fxName, fxPos);
				
				int lastHealth = actor.Health();
				actor.InflictGenericDamage(playerActor, damage);
				
				if (actor.ScriptObject() !is null && actor.ScriptObject().obj !is null)
				{
					TurokEnemy@ enemyObj = cast<TurokEnemy@>(actor.ScriptObject().obj);
					if (lastHealth > 0 && actor.Health() <= 0 && enemyObj !is null)
					{
						enemyObj.m_bMortallyWounded = true;
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
