//------------------------------------------------------------------------------------------------------------------------
// Class       : BP::CL_BP_PickupInventory
// Author      : BehemothProgrammer
// Description : When the player enters the actors radius this item will be picked up and added to the inventory and this can trigger other actors or map scripts.
// OnTrigger   : If is not marked persistent then will hide/show the item.
//
// Parameters  : (7) Params 1 -128..127 				: Inventory Item Index 0..127
//               (3) Max Regenerations -32768..32767	: Flags
//               (0) Target TID -32768..32767 			: Actors TID to trigger on touch. If is less than 1 this will not be used.
//               (2) Health -32768..32767				: Script to trigger on touch. If is less than 1 this will not be used.
//               StepHeight/DeadHeight					: Sound ID to play on pickup if greater than 0 (ie. 389 generic_4_non_weapon_pickup)
//------------------------------------------------------------------------------------------------------------------------
namespace BP
{
	enum PickupInventoryFlags
	{
		PINVENTORY_FLAG_STARTHIDDEN = (1 << 0),
		PINVENTORY_FLAG_NOSPIN = (1 << 1),
		PINVENTORY_FLAG_NOPERSIST = (1 << 2),
		PINVENTORY_FLAG_SPARKLES = (1 << 3),
		PINVENTORY_FLAG_SPAWNFX = (1 << 4), //plays spawnfx when shown
		PINVENTORY_FLAG_CUSTOMSOUND = (1 << 5), //use custom sounds from BP_GameDefCustom actor def
	}

	class CL_BP_PickupInventory : CL_BP_Pickup
	{
		int inventoryIndex;
		int flags;
		int triggerTID;
		int triggerScriptID;
		int pickupSound;
		
		bool isActive;
		//------------------------------------------------------------------------------------------------------------------------
		CL_BP_PickupInventory(kActor @actor)
		{
			super(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnBeginLevel()
		{
			inventoryIndex = self.SpawnParams(7);
			flags = self.SpawnParams(3);
			triggerTID = self.SpawnParams(0);
			triggerScriptID = self.SpawnParams(2);
			pickupSound = int(self.StepHeight());

			self.Flags() &= ~AF_SOLID;
			self.Flags() |= AF_INVINCIBLE;
			
			if (self.IsPersistentMarked())
			{
				self.Remove();
			}
			else
			{
				SetActive((flags & PINVENTORY_FLAG_STARTHIDDEN == 0));
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			if (!isActive)
				return;

			if (flags & PINVENTORY_FLAG_NOSPIN == 0)
				self.Yaw() += BP::Math::ONEDEGREERAD;
			
			if (flags & PINVENTORY_FLAG_SPARKLES != 0 && self.GameTicks() % 2 == 0)
				self.SpawnFx("fx/generic_230.kfx", kVec3(0, 0, GAME_SCALE));
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTouch(kActor @actor)
		{
			if (actor.Type() != AT_PLAYER) return;
			
			OnPickup(actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnActivate()
		{
			self.Flags() &= ~AF_ACTIVATED;
			
			SetActive(!isActive);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetActive(const bool active)
		{
			isActive = active;
			if (isActive)
			{
				self.Flags() |= AF_CANBETOUCHED;
				self.Flags() &= ~AF_NODRAW;
				
				//don't do spawn effects/sound if spawned at start
				if (flags & PINVENTORY_FLAG_SPAWNFX != 0 && self.GameTicks() != 0)
				{
					kVec3 org = self.Origin();
					org.z += (3 * GAME_SCALE);
					self.PlaySound("sounds/shaders/generic_3_energy_pickup.ksnd");
					Game.SpawnFx("fx/generic_260.kfx", org, self.SectorIndex());
				}
			}
			else
			{
				self.Flags() &= ~AF_CANBETOUCHED;
				self.Flags() |= AF_NODRAW;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnPickup(kActor @actor)
		{
			if (flags & PINVENTORY_FLAG_NOPERSIST == 0)
				self.MarkPersistentBit(false);
			
			BP::Inventory::Give(inventoryIndex);
			
			if (pickupSound > 0)
			{
				if (flags & PINVENTORY_FLAG_CUSTOMSOUND != 0)
					self.PlaySound(BP::Game::GetCustomSound(pickupSound));
				else
					self.PlaySoundWithLookupID(pickupSound);
			}
			if (triggerTID > 0)
				World.TriggerActorsByTID(self, triggerTID);
			if (triggerScriptID > 0)
				Game.CallDelayedMapScript(triggerScriptID, self, 0.0f);
			self.Remove();
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
