namespace BP { namespace Actor
{
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const int id, kActor@ destActor)
	{
		return Spawn(id, destActor.Origin(), destActor.Yaw(), destActor.SectorIndex());
	}
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const int id, const kVec3 &in origin, const float yaw = 0.0f, const int sector = -1)
	{
		return Spawn(id, origin.x, origin.y, origin.z, yaw, sector);
	}
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const int id, const float x = 0.0f, const float y = 0.0f, const float z = 0.0f, const float yaw = 0.0f, const int sector = -1)
	{
		kActor@ actor = ActorFactory.Spawn(id, x, y, z, yaw, sector);
		ProcessPostSpawn(actor);
		return actor;
	}
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const kStr &in name, kActor@ destActor)
	{
		return Spawn(name, destActor.Origin(), destActor.Yaw(), destActor.SectorIndex());
	}
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const kStr &in name, const kVec3 &in origin, const float yaw = 0.0f, const int sector = -1)
	{
		return Spawn(name, origin.x, origin.y, origin.z, yaw, sector);
	}
	//------------------------------------------------------------------------------------------------------------------------
	kActor@ Spawn(const kStr &in name, const float x = 0.0f, const float y = 0.0f, const float z = 0.0f, const float yaw = 0.0f, const int sector = -1)
	{
		kActor@ actor = ActorFactory.Spawn(name, x, y, z, yaw, sector);
		ProcessPostSpawn(actor);
		return actor;
	}
	//------------------------------------------------------------------------------------------------------------------------
	void ProcessPostSpawn(kActor@ actor)
	{
		if (!Exists(actor))
			return;
		
		if (GetDefBool(actor, DEF_SOLID))
			actor.Flags() |= AF_SOLID;
	}
	//------------------------------------------------------------------------------------------------------------------------
	array<kActor@> Find(const array<int> &in tids)
	{
		array<kActor@> actors;
		for (uint i = 0; i < tids.length(); i++)
		{
			kActor@ actor = World.GetActorByTID(tids[i]);
			if (Exists(actor))
				actors.insertLast(actor);
		}
		
		return actors;
	}
	//------------------------------------------------------------------------------------------------------------------------
	array<kActor@> Find(const int startTID, const int endTID)
	{
		array<kActor@> actors;
		for (int i = startTID; i <= endTID; i++)
		{
			kActor@ actor = World.GetActorByTID(i);
			if (Exists(actor))
				actors.insertLast(actor);
		}
		
		return actors;
	}
	//------------------------------------------------------------------------------------------------------------------------
	void SetPosition(kActor@ actor, kActor@ destActor, const bool floorHeight = false)
	{
		actor.Origin() = destActor.Origin();
		actor.SetSector(destActor.SectorIndex());
		if (floorHeight)
			actor.Origin().z = destActor.FloorHeight();
	}
	//------------------------------------------------------------------------------------------------------------------------
	void TossItem(kActor@ actor, const float velScale = 1.0f)
	{
		actor.Scale().Set(0.35f, 0.35f, 0.35f);
		actor.ClipFlags() = (CF_DROPOFF|CF_CLIPEDGES|CF_NOCLIPACTORS|CF_COLLIDEFLOORS|CF_COLLIDEHEIGHT);
		actor.Velocity() = kVec3(0, 0, 1).Randomize(0.25f).Normalize() * 200.0f * velScale;
		actor.Velocity() *= (1.0f / 60.0f);
		actor.BounceDamp() = 0.5f;
		actor.Gravity() = 0.5f;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Useful to call custom functions from the map script module to the game script module.
	// Otherwise just use the ScriptObject's ScriptObject() function to get a pointer to it.
	// 
	// Required:
	// Function header: void funcName(kActor @actor, const float arg1, const float arg2, const float arg3, const float arg4)
	// Function body condition check: if (actor.Type() != AT_FUNCTIONCALL) return;
	//------------------------------------------------------------------------------------------------------------------------
	void CallFunc(kActor@ actor, const kStr &in funcName, const float arg1 = 0.0f, const float arg2 = 0.0f, const float arg3 = 0.0f, const float arg4 = 0.0f)
	{
		kActor@ playerActor = Player.Actor().CastToActor();
		kActor@ dummy = Spawn(AT_FUNCTIONCALL, playerActor.Origin(), 0.0f, playerActor.SectorIndex());
		dummy.Flags() |= AF_SOLID;
		actor.InteractActorsAtPosition(playerActor.Origin(), funcName, arg1, arg2, arg3, arg4);
		dummy.Remove();
	}
	//------------------------------------------------------------------------------------------------------------------------
	void TogglePersistentBit(kActor@ actor)
	{
		actor.MarkPersistentBit(actor.IsPersistentMarked());
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsDead(kActor@ actor)
	{
		return ((actor.Flags() & AF_DEAD) != 0);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsInWater(kActor@ actor)
	{
		int waterLevel = actor.GetWaterLevel();
		return (waterLevel == WLT_BETWEEN || waterLevel == WLT_UNDER);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsHidden(kActor@ actor)
	{
		return ((actor.Flags() & AF_HIDDEN) != 0);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool Exists(kActor@ actor)
	{
		return @actor !is null && !actor.IsStale();
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool EnteredArea(kActor@ actor)
	{
		return ((actor.Flags() & AF_ENTEREDAREAEVENT) == 0);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsInSaveArea(kActor@ actor)
	{
		return (World.GetAreaFlags(actor.AreaID()) & AAF_SAVEGAME) != 0 && actor.OnGround();
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsClimbing(kActor@ actor)
	{
		return (World.GetAreaFlags(actor.AreaID()) & AAF_CLIMB) != 0 && actor.OnGround();
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsInRange(kActor@ actor, kActor@ actor2, const float range)
	{
		return actor.Origin().DistanceSq(actor2.Origin()) < range * range;
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsInRange(kActor@ actor, const kVec3 &in pos, const float range)
	{
		return actor.Origin().DistanceSq(pos) < range * range;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Return true if actors collision cylinder is inside the pos
	//------------------------------------------------------------------------------------------------------------------------
	bool InPoint(kActor @actor, const kVec3 &in pos)
	{
		kVec3 p = actor.Origin();
		return IsInRange(actor, pos, actor.Radius()) && pos.z >= p.z && pos.z <= p.z + actor.Height();
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool InActorView(kActor @actor, kActor @actor2, const float angle)
	{
		return Math::Fabs(actor.GetTurnYaw(actor2.Origin())) < angle;
	}
	//------------------------------------------------------------------------------------------------------------------------
	void HideAI(kActor@ actor)
	{
		actor.Flags() |= AF_HIDDEN | AF_DISABLED | AF_INVINCIBLE;
	}
	//------------------------------------------------------------------------------------------------------------------------
	void ShowAI(kActor@ actor)
	{
		actor.Flags() &= ~(AF_HIDDEN | AF_DISABLED | AF_INVINCIBLE);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsDisabled(kActor@ actor)
	{
		return (actor.Flags() & AF_DISABLED) != 0;
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool IsActivated(kActor@ actor)
	{
		return (actor.Flags() & AF_ACTIVATED) != 0;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns the forward direction of the actor
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 ForwardDir(kActor@ actor)
	{
		return kVec3(0.0f, 1.0f, 0.0f) * actor.Rotation();
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns the right direction of the actor
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 RightDir(kActor@ actor)
	{
		return kVec3(1.0f, 0.0f, 0.0f) * actor.Rotation();
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns the up direction of the actor
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 UpDir(kActor@ actor)
	{
		return kVec3(0.0f, 0.0f, 1.0f) * actor.Rotation();
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool GetDefBool(kActor@ actor, const kStr &in key, const bool defaultValue = false)
	{
		kDictMem@ def = actor.Definition();
		bool result;
		if (def !is null && def.GetBool(key, result))
			return result;
		
		return defaultValue;
	}
	//------------------------------------------------------------------------------------------------------------------------
	int GetDefInt(kActor@ actor, const kStr &in key, const int defaultValue = 0)
	{
		kDictMem@ def = actor.Definition();
		int result;
		if (def !is null && def.GetInt(key, result))
			return result;
		
		return defaultValue;
	}
	//------------------------------------------------------------------------------------------------------------------------
	kStr GetDefString(kActor@ actor, const kStr &in key, const kStr &in defaultValue = "")
	{
		kDictMem@ def = actor.Definition();
		kStr result;
		if (def !is null && def.GetString(key, result))
			return result;
		
		return defaultValue;
	}
	//------------------------------------------------------------------------------------------------------------------------
	array<kStr> GetDefStringArray(kActor@ actor, kStr &in key)
	{
		array<kStr> result;
		kDictMem@ def = actor.Definition();
		if (def !is null)
		{
			int i = 1;
			kStr s = "";
			kStr fullKey = key + "_";
			while (true)
			{
				if (def.GetString(fullKey + i, s))
					result.insertLast(s);
				else
					break;
				i++;
			}
		}
		
		return result;
	}
	//------------------------------------------------------------------------------------------------------------------------
}}
