namespace BP { namespace Math
{
	const uint UBIT31 = 2147483648;
	const int MIN_INT = -2147483648;
	const int MAX_INT = 2147483647;
	const float ONEDEGREERAD = 0.017453f;
	const float EPSILON = 1.401298E-45;
	const float PI_HALF = 1.570796f;
	//------------------------------------------------------------------------------------------------------------------------
	bool InRange(const int value, const int min, const int max)
	{
		return value >= min && value <= max;
	}
	//------------------------------------------------------------------------------------------------------------------------
	float Min(const float a, const float b)
	{
		return (a < b) ? a : b;
	}
	//------------------------------------------------------------------------------------------------------------------------
	float Max(const float a, const float b)
	{
		return (a > b) ? a : b;
	}
	//------------------------------------------------------------------------------------------------------------------------
	int Min(const int a, const int b)
	{
		return (a < b) ? a : b;
	}
	//------------------------------------------------------------------------------------------------------------------------
	int Max(const int a, const int b)
	{
		return (a > b) ? a : b;
	}
	//------------------------------------------------------------------------------------------------------------------------
	float SmoothStepFloat(const float from, const float to, float t)
	{
		t = Clampf(t, 0.0f, 1.0f);
		t = -2.0f * t * t * t + 3.0f * t * t;
		return to * t + from * (1.0f - t);
	}
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 SmoothStepVec3(const kVec3 &in from, const kVec3 &in to, float t)
	{
		t = -2.0f * t * t * t + 3.0f * t * t;
		float it = 1.0f - t;
		return kVec3(to.x * t + from.x * it, to.y * t + from.y * it, to.z * t + from.z * it);
	}
	//------------------------------------------------------------------------------------------------------------------------
	float LerpClamped(const float start, const float end, float time)
	{
		return Math::Lerp(start, end, Clampf(time, 0.0f, 1.0f));
	}
	//------------------------------------------------------------------------------------------------------------------------
	int LerpInt(const int start, const int end, const float time)
	{
		return int(Math::Lerp(start, end, time));
	}
	//------------------------------------------------------------------------------------------------------------------------
	int LerpClampedInt(const int start, const int end, const float time)
	{
		return int(Math::Lerp(start, end, Clampf(time, 0.0f, 1.0f)));
	}
	//------------------------------------------------------------------------------------------------------------------------
	float Clampf(const float value, const float min, const float max)
	{
		return (value < min) ? min : ((value > max) ? max : value);
	}
	//------------------------------------------------------------------------------------------------------------------------
	int Clamp(const int value, const int min, const int max)
	{
		return (value < min) ? min : ((value > max) ? max : value);
	}
	//------------------------------------------------------------------------------------------------------------------------
	float Round(const float value)
	{
		int a = int(value * 10);
		int b = int(value) * 10;
		if (value < 0)
		{
			return (a + b >= 5) ? Math::Floor(value) : Math::Ceil(value);
		}
		return (a - b >= 5) ? Math::Ceil(value) : Math::Floor(value);
	}
	//------------------------------------------------------------------------------------------------------------------------
	int RoundToInt(const float value)
	{
		return int(Round(value));
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns a clamped angle to intervals of angleClamp degrees
	//------------------------------------------------------------------------------------------------------------------------
	float ClampAngle(const float value, const float angleClamp)
	{
		return Round(value / angleClamp) * angleClamp;
	}
	//------------------------------------------------------------------------------------------------------------------------
	float Vec3Dot(const kVec3 &in lhs, const kVec3 &in rhs)
	{
		return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z;
	}
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 Vec3Reflect(const kVec3 &in inDirection, const kVec3 &in inNormal)
	{
		//missing add and mul operators so doing addassign and mulassign instead
		float t2 = -2.0f * Vec3Dot(inNormal, inDirection);
		kVec3 t = inNormal;
		t *= t2;
		t += inDirection;
		return t;
		//return -2.0f * Vec3Dot(inNormal, inDirection) * inNormal + inDirection;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns the number of bits from startbit to startbit + count that are on
	//------------------------------------------------------------------------------------------------------------------------
	int BitCount(const int value, const int startBit, const int count = 1)
	{
		int result = 0;
		for (int i = startBit; i < startBit + count; i++)
		{
			if ((value & (1 << i)) != 0)
			{
				result++;
			}
		}
		return result;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns an array of digits from the integer
	//------------------------------------------------------------------------------------------------------------------------
	array<int> Digits(const uint n, const bool reverse = false)
	{
		if (n == 0)
			return array<int> = {0};
		
		uint num = n;
		array<int> digits;
		while (num > 0)
		{
			digits.insertLast(num % 10);
			num /= 10;
		}
		if (!reverse)
			digits.reverse();
		return digits;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Returns the number of digits in the integer (ignores minus sign for negative values)
	//------------------------------------------------------------------------------------------------------------------------
	int DigitCount(const int n)
	{
		int num = Math::Abs(n);
		int count = 0;
		while (num > 0)
		{
			num /= 10;
			count++;
		}
		return count;
	}
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 RandPointInCircle()
	{
		float theta = Math::RandRange(0.0f, 2.0f * Math::pi);
		float distance = Math::Sqrt(Math::RandFloat());
		return kVec3(distance * Math::Cos(theta), distance * Math::Sin(theta), 0.0f);
	}
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 RandPointInCircle(const kVec3 &in center, const float diameter)
	{
		return (RandPointInCircle() * diameter) + center;
	}
	//------------------------------------------------------------------------------------------------------------------------
	kVec3 RandPointInTriangle(kVec3 &in A, kVec3 &in B, kVec3 &in C)
	{
		// float r1 = Math::RandRange(0.0f, 1.0f);
		// float r2 = Math::RandRange(0.0f, 1.0f);
		// float sqrR1 = Math::Sqrt(r1);
		// float vX = (1 - sqrR1) * a.x + (sqrR1 * (1 - r2)) * b.x + (sqrR1 * r2) * c.x;
		// float vY = (1 - sqrR1) * a.y + (sqrR1 * (1 - r2)) * b.y + (sqrR1 * r2) * c.y;
		
		float x = Math::RandRange(0.0f, 1.0f);
		float y = Math::RandRange(0.0f, 1.0f);
		if (x + y > 1)
		{
			x = 1 - x;
			y = 1 - y;
		}
		float d = 1-x-y;
		float vX = (x * A.x) + (y * B.x) + (d * C.x);
		float vY = (x * A.y) + (y * B.y) + (d * C.y);
		
		return kVec3(vX, vY, 0.0f);
	}
	//------------------------------------------------------------------------------------------------------------------------
	float AreaOfTriangle(kVec3 &in A, kVec3 &in B, kVec3 &in C)
	{
		float valueA = (A - B).Unit();
		float valueB = (C - B).Unit();
		float valueC = (A - C).Unit();
		
		float i = (valueA + valueB + valueC) * 0.5f;
		return Math::Sqrt(i * (i - valueA) * (i - valueB) * (i - valueC));
	}
	//------------------------------------------------------------------------------------------------------------------------
	array<int> Triangulate(const array<kVec3> &points)
	{
		array<int> indices;
		
		int n = int(points.length());
		if (n < 3)
			return indices;
		
		array<int> V(n);
		if (TriangulateArea(points) > 0.0f)
		{
			for (int v = 0; v < n; v++)
				V[v] = v;
		}
		else
		{
			for (int v = 0; v < n; v++)
				V[v] = (n - 1) - v;
		}
		
		int nv = n;
		int count = 2 * nv;
		for (int v = nv - 1; nv > 2; )
		{
			if ((count--) <= 0)
			{
				return indices;
			}
 
			int u = v;
			if (nv <= u)
				u = 0;
			v = u + 1;
			if (nv <= v)
				v = 0;
			int w = v + 1;
			if (nv <= w)
				w = 0;
 
			if (TriangulateSnip(points, u, v, w, nv, V))
			{
				int a, b, c;
				a = V[u];
				b = V[v];
				c = V[w];
				indices.insertLast(a);
				indices.insertLast(b);
				indices.insertLast(c);
				for (int s = v; s + 1 < nv; s++)
					V[s] = V[s + 1];
				nv--;
				count = 2 * nv;
				
			}
		}
 
		indices.reverse();
		return indices;
	}
	//------------------------------------------------------------------------------------------------------------------------
	float TriangulateArea(const array<kVec3> &points)
	{
		int n = (points.length());
		float A = 0.0f;
		int q = 0;
		for (int p = n - 1; q < n; p = q++)
		{
			kVec3 pval = points[p];
			kVec3 qval = points[q];
			A += pval.x * qval.y - qval.x * pval.y;
		}
		return (A * 0.5f);
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool TriangulateSnip(const array<kVec3> &points, const int u, const int v, const int w, const int n, const array<int> &in V)
	{
		int p;
		kVec3 A = points[V[u]];
		kVec3 B = points[V[v]];
		kVec3 C = points[V[w]];
		if (EPSILON > (((B.x - A.x) * (C.y - A.y)) - ((B.y - A.y) * (C.x - A.x))))
			return false;
		for (p = 0; p < n; p++) {
			if ((p == u) || (p == v) || (p == w))
				continue;
			kVec3 P = points[V[p]];
			if (TriangulateInsideTriangle(A, B, C, P))
				return false;
		}
		return true;
	}
	//------------------------------------------------------------------------------------------------------------------------
	bool TriangulateInsideTriangle(const kVec3 &in A, const kVec3 &in B, const kVec3 &in C, const kVec3 &in P)
	{
		float ax, ay, bx, by, cx, cy, apx, apy, bpx, bpy, cpx, cpy;
		float cCROSSap, bCROSScp, aCROSSbp;
 
		ax = C.x - B.x; ay = C.y - B.y;
		bx = A.x - C.x; by = A.y - C.y;
		cx = B.x - A.x; cy = B.y - A.y;
		apx = P.x - A.x; apy = P.y - A.y;
		bpx = P.x - B.x; bpy = P.y - B.y;
		cpx = P.x - C.x; cpy = P.y - C.y;
 
		aCROSSbp = ax * bpy - ay * bpx;
		cCROSSap = cx * apy - cy * apx;
		bCROSScp = bx * cpy - by * cpx;
 
		return ((aCROSSbp >= 0.0f) && (bCROSScp >= 0.0f) && (cCROSSap >= 0.0f));
	}
	//------------------------------------------------------------------------------------------------------------------------
}}
