//------------------------------------------------------------------------------------------------------------------------
// Class       : Q::CL_Q_Map
// Author      : BehemothProgrammer
// Description : Map manager actor that should be spawned in via script only and never placed in the map itself with the editor.
// OnTrigger   : Nothing.
//
// Parameters  : 
//------------------------------------------------------------------------------------------------------------------------
namespace Q
{
	float biosuitTime; //Time when biosuit will become inactive
	bool biosuitFading;
	bool hasBiosuit;

	//Special Areas have the Flag "Toggle" on and Arg 1 is the flags that define what special areas it is.
	const uint SAREA_TOXICWATER = (1 << 0); //Arg 5: damage amount, Arg 6: Damage time in seconds / 10)
	const uint SAREA_SECRET = (1 << 1); //Arg 6: Secret Index (0..31)
	const uint SAREA_WATERFLOOR = (1 << 2); //Arg 2: Water Floor Height. (Sets The water flag correctly. For water floating in air.)

	const int SAREAFLAG_COUNT = 3; //number of special area flags used

	class CL_Q_Map : BP::CL_BP_ScriptObject
	{
		int lastSpecialAreaFlags;
		int specialAreaFlags; //from area arg (16 bit)
		int playerAreaID;
		int lastAreaID;
		int playerSectorIndex;
		int lastSectorIndex;
		uint lastAreaFlags;
		uint curAreaFlags;
		float toxicWaterTime; //Amount of time player has been in toxic water
		//------------------------------------------------------------------------------------------------------------------------
		CL_Q_Map(kActor @actor)
		{
			super(actor);
			biosuitTime = 0.0f;
			biosuitFading = false;
			hasBiosuit = false;
			toxicWaterTime = 0.0f;
			playerAreaID = 0;
			lastSpecialAreaFlags = 0;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTick() override
		{
			//tick biosuit time
			if (hasBiosuit)
			{
				if (!biosuitFading)
				{
					if (BP::Time::Reached(biosuitTime - 3.0f))
					{
						biosuitFading = true;
						BP::HUD::ShowMessage(BP::Game::GetLocalText(11), 240);
						BP::playerActor.PlaySound(PATH_SOUND_SUIT2);
					}
				}
				else
				{
					if (BP::Time::Reached(biosuitTime))
					{
						hasBiosuit = false;
					}
				}
			}
			
			OnTickSpecialSectors();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnEndLevel()
		{
			float mapStartTime = BP::GameVar::GetFloat(GV_KEY_MAP_TIME);
			int mapSecrets = BP::GameVar::GetInt(GV_KEY_MAP_SECRETS);
			int mapKills = BP::GameVar::GetInt(GV_KEY_MAP_KILLS);
			
			//if did exit map then use exit time for time
			float mapExitTotalTime = BP::GameVar::GetFloat(GV_KEY_MAP_EXITTIME);
			if (mapExitTotalTime <= 0.0f)
				mapExitTotalTime = mapStartTime + BP::Time::MapTime();
			
			SetMapData(Game.GetCurrentMapID(), mapExitTotalTime, mapSecrets, mapKills);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnEnterSpecialArea(const uint specialAreaFlag)
		{
			switch (specialAreaFlag)
			{
				case SAREA_SECRET:
				{
					int secretIndex = World.GetAreaArg(playerAreaID, 5);
					FindSecret(secretIndex);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnExitSpecialArea(const uint specialAreaFlag)
		{
			switch (specialAreaFlag)
			{
				case SAREA_WATERFLOOR:
				{
					if (lastAreaFlags & AAF_WATER != 0)
					{
						World.FloodMatchingAreaFlags(lastSectorIndex, AAF_WATER, false);
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTickSpecialArea(const uint specialAreaFlag)
		{
			switch (specialAreaFlag)
			{
				case SAREA_TOXICWATER:
				{
					int waterLevel = BP::playerActor.GetWaterLevel();
					if (toxicWaterTime <= 0.0f && !hasBiosuit && (waterLevel == WLT_BETWEEN || waterLevel == WLT_UNDER))
					{
						toxicWaterTime = float(World.GetAreaArg(playerAreaID, 5)) / 10.0f;
						int damage = World.GetAreaArg(playerAreaID, 4);
						BP::playerActor.InflictGenericDamage(self, damage);
					}
					break;
				}
				case SAREA_WATERFLOOR:
				{
					float waterFloorHeight = float(World.GetAreaArg(playerAreaID, 1)) - BP::playerActor.Height();
					if (BP::playerActor.Origin().z >= waterFloorHeight)
					{
						if (curAreaFlags & AAF_WATER == 0)
						{
							World.FloodMatchingAreaFlags(playerSectorIndex, AAF_WATER, true);
						}
					}
					else
					{
						if (curAreaFlags & AAF_WATER != 0)
						{
							World.FloodMatchingAreaFlags(playerSectorIndex, AAF_WATER, false);
						}
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnTickSpecialSectors()
		{
			uint sflag = 0;
			//Tick Player on Special Area Sectors
			playerAreaID = BP::playerActor.AreaID();
			playerSectorIndex = BP::playerActor.SectorIndex();
			specialAreaFlags = 0;
			curAreaFlags = World.GetAreaFlags(playerAreaID);
			if (curAreaFlags & AAF_TOGGLE != 0)
				specialAreaFlags = World.GetAreaArg(playerAreaID, 0);
			
			toxicWaterTime -= GAME_DELTA_TIME;
			
			//Player Changed Special Area Sectors
			if (lastSpecialAreaFlags != specialAreaFlags)
			{
				for (uint i = 0; i < SAREAFLAG_COUNT; i++)
				{
					sflag = (1 << i);
					if (lastSpecialAreaFlags & sflag != 0 && specialAreaFlags & sflag == 0)
						OnExitSpecialArea(sflag);
				}
				for (uint i = 0; i < SAREAFLAG_COUNT; i++)
				{
					sflag = (1 << i);
					if (specialAreaFlags & sflag != 0 && lastSpecialAreaFlags & sflag == 0)
						OnEnterSpecialArea(sflag);
				}
				
				lastSpecialAreaFlags = specialAreaFlags;
			}
			
			for (uint i = 0; i < SAREAFLAG_COUNT; i++)
			{
				sflag = (1 << i);
				if (specialAreaFlags & sflag == 0)
					continue;
				
				OnTickSpecialArea(sflag);
			}
			
			lastAreaFlags = curAreaFlags;
			lastAreaID = playerAreaID;
			lastSectorIndex = playerSectorIndex;
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}